# -*- coding: utf-8 -*-

# Import python libs
from __future__ import absolute_import

import os
import sys

import salt.modules.pip as pip

# Import salt libs
import salt.utils.platform
from salt.exceptions import CommandExecutionError

# Import Salt Testing libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class PipTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {pip: {"__salt__": {"cmd.which_bin": lambda _: "pip"}}}

    def test__pip_bin_env(self):
        ret = pip._pip_bin_env(None, "C:/Users/ch44d/Documents/salt/tests/pip.exe")
        if salt.utils.platform.is_windows():
            self.assertEqual(ret, "C:/Users/ch44d/Documents/salt/tests")
        else:
            self.assertIsNone(ret)

    def test__pip_bin_env_no_change(self):
        cwd = "C:/Users/ch44d/Desktop"
        ret = pip._pip_bin_env(cwd, "C:/Users/ch44d/Documents/salt/tests/pip.exe")
        self.assertEqual(ret, cwd)

    def test__pip_bin_env_no_bin_env(self):
        ret = pip._pip_bin_env(None, None)
        self.assertIsNone(ret)

    def test_install_frozen_app(self):
        pkg = "pep8"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch("sys.frozen", True, create=True):
            with patch("sys._MEIPASS", True, create=True):
                with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                    pip.install(pkg)
                    expected = [
                        sys.executable,
                        "pip",
                        "install",
                        pkg,
                    ]
                    mock.assert_called_with(
                        expected,
                        python_shell=False,
                        saltenv="base",
                        use_vt=False,
                        runas=None,
                    )

    def test_install_source_app(self):
        pkg = "pep8"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch("sys.frozen", False, create=True):
            with patch("sys._MEIPASS", False, create=True):
                with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                    pip.install(pkg)
                    expected = [
                        sys.executable,
                        "-m",
                        "pip",
                        "install",
                        pkg,
                    ]
                    mock.assert_called_with(
                        expected,
                        python_shell=False,
                        saltenv="base",
                        use_vt=False,
                        runas=None,
                    )

    def test_fix4361(self):
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(requirements="requirements.txt")
            expected_cmd = [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--requirement",
                "requirements.txt",
            ]
            mock.assert_called_with(
                expected_cmd,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

    def test_install_editable_without_egg_fails(self):
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            self.assertRaises(
                CommandExecutionError,
                pip.install,
                editable="git+https://github.com/saltstack/salt-testing.git",
            )

    def test_install_multiple_editable(self):
        editables = [
            "git+https://github.com/jek/blinker.git#egg=Blinker",
            "git+https://github.com/saltstack/salt-testing.git#egg=SaltTesting",
        ]

        expected = [sys.executable, "-m", "pip", "install"]
        for item in editables:
            expected.extend(["--editable", item])

        # Passing editables as a list
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(editable=editables)
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

        # Passing editables as a comma separated list
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(editable=",".join(editables))
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

    def test_install_multiple_pkgs_and_editables(self):
        pkgs = ["pep8", "salt"]
        editables = [
            "git+https://github.com/jek/blinker.git#egg=Blinker",
            "git+https://github.com/saltstack/salt-testing.git#egg=SaltTesting",
        ]

        expected = [sys.executable, "-m", "pip", "install"]
        expected.extend(pkgs)
        for item in editables:
            expected.extend(["--editable", item])

        # Passing editables as a list
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkgs=pkgs, editable=editables)
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

        # Passing editables as a comma separated list
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkgs=",".join(pkgs), editable=",".join(editables))
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

        # As single string (just use the first element from pkgs and editables)
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkgs=pkgs[0], editable=editables[0])
            expected = [
                sys.executable,
                "-m",
                "pip",
                "install",
                pkgs[0],
                "--editable",
                editables[0],
            ]
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

    def test_issue5940_install_multiple_pip_mirrors(self):
        """
        test multiple pip mirrors.  This test only works with pip < 7.0.0
        """
        with patch.object(pip, "version", MagicMock(return_value="1.4")):
            mirrors = [
                "http://g.pypi.python.org",
                "http://c.pypi.python.org",
                "http://pypi.crate.io",
            ]

            expected = [sys.executable, "-m", "pip", "install", "--use-mirrors"]
            for item in mirrors:
                expected.extend(["--mirrors", item])
            expected.append("pep8")

            # Passing mirrors as a list
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                pip.install(pkgs=["pep8"], mirrors=mirrors)
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

            # Passing mirrors as a comma separated list
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                pip.install(pkgs=["pep8"], mirrors=",".join(mirrors))
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

            expected = [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--use-mirrors",
                "--mirrors",
                mirrors[0],
                "pep8",
            ]

            # As single string (just use the first element from mirrors)
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                pip.install(pkgs=["pep8"], mirrors=mirrors[0])
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

    def test_install_with_multiple_find_links(self):
        find_links = [
            "http://g.pypi.python.org",
            "http://c.pypi.python.org",
            "http://pypi.crate.io",
        ]
        pkg = "pep8"

        expected = [sys.executable, "-m", "pip", "install"]
        for item in find_links:
            expected.extend(["--find-links", item])
        expected.append(pkg)

        # Passing mirrors as a list
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, find_links=find_links)
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

        # Passing mirrors as a comma separated list
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, find_links=",".join(find_links))
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

        # Valid protos work?
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, find_links=find_links)
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

        expected = [
            sys.executable,
            "-m",
            "pip",
            "install",
            "--find-links",
            find_links[0],
            pkg,
        ]

        # As single string (just use the first element from find_links)
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, find_links=find_links[0])
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

        # Invalid proto raises exception
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            self.assertRaises(
                CommandExecutionError,
                pip.install,
                "'" + pkg + "'",
                find_links="sftp://pypi.crate.io",
            )

    def test_install_no_index_with_index_url_or_extra_index_url_raises(self):
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            self.assertRaises(
                CommandExecutionError,
                pip.install,
                no_index=True,
                index_url="http://foo.tld",
            )

        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            self.assertRaises(
                CommandExecutionError,
                pip.install,
                no_index=True,
                extra_index_url="http://foo.tld",
            )

    def test_install_failed_cached_requirements(self):
        with patch(
            "salt.modules.pip._get_cached_requirements"
        ) as get_cached_requirements:
            get_cached_requirements.return_value = False
            ret = pip.install(requirements="salt://my_test_reqs")
            self.assertEqual(False, ret["result"])
            self.assertIn("my_test_reqs", ret["comment"])

    def test_install_cached_requirements_used(self):
        with patch(
            "salt.modules.pip._get_cached_requirements"
        ) as get_cached_requirements:
            get_cached_requirements.return_value = "my_cached_reqs"
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                pip.install(requirements="salt://requirements.txt")
                expected = [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "--requirement",
                    "my_cached_reqs",
                ]
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

    def test_install_venv(self):
        with patch("os.path") as mock_path:

            def join(*args):
                return os.path.normpath(os.sep.join(args))

            mock_path.is_file.return_value = True
            mock_path.isdir.return_value = True
            mock_path.join = join

            if salt.utils.platform.is_windows():
                venv_path = "C:\\test_env"
                bin_path = os.path.join(venv_path, "python.exe")
            else:
                venv_path = "/test_env"
                bin_path = os.path.join(venv_path, "python")

            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            pip_bin = MagicMock(return_value=[bin_path, "-m", "pip"])

            with patch.dict(pip.__salt__, {"cmd.run_all": mock}), patch.object(
                pip, "_get_pip_bin", pip_bin
            ):
                pip.install("mock", bin_env=venv_path)
                mock.assert_called_with(
                    [bin_path, "-m", "pip", "install", "mock"],
                    env={"VIRTUAL_ENV": venv_path},
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

    def test_install_log_argument_in_resulting_command(self):
        with patch("os.access") as mock_path:
            pkg = "pep8"
            log_path = "/tmp/pip-install.log"
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                pip.install(pkg, log=log_path)
                expected = [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "--log",
                    log_path,
                    pkg,
                ]
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

    def test_non_writeable_log(self):
        with patch("os.path") as mock_path:
            # Let's fake a non-writable log file
            pkg = "pep8"
            log_path = "/tmp/pip-install.log"
            mock_path.exists.side_effect = IOError("Fooo!")
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                self.assertRaises(IOError, pip.install, pkg, log=log_path)

    def test_install_timeout_argument_in_resulting_command(self):
        # Passing an int
        pkg = "pep8"
        expected = [sys.executable, "-m", "pip", "install", "--timeout"]
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, timeout=10)
            mock.assert_called_with(
                expected + [10, pkg],
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

        # Passing an int as a string
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, timeout="10")
            mock.assert_called_with(
                expected + ["10", pkg],
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

        # Passing a non-int to timeout
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            self.assertRaises(ValueError, pip.install, pkg, timeout="a")

    def test_install_index_url_argument_in_resulting_command(self):
        pkg = "pep8"
        index_url = "http://foo.tld"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, index_url=index_url)
            expected = [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--index-url",
                index_url,
                pkg,
            ]
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

    def test_install_extra_index_url_argument_in_resulting_command(self):
        pkg = "pep8"
        extra_index_url = "http://foo.tld"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, extra_index_url=extra_index_url)
            expected = [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--extra-index-url",
                extra_index_url,
                pkg,
            ]
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

    def test_install_no_index_argument_in_resulting_command(self):
        pkg = "pep8"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, no_index=True)
            expected = [sys.executable, "-m", "pip", "install", "--no-index", pkg]
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

    def test_install_build_argument_in_resulting_command(self):
        pkg = "pep8"
        build = "/tmp/foo"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, build=build)
            expected = [sys.executable, "-m", "pip", "install", "--build", build, pkg]
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

    def test_install_target_argument_in_resulting_command(self):
        pkg = "pep8"
        target = "/tmp/foo"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, target=target)
            expected = [sys.executable, "-m", "pip", "install", "--target", target, pkg]
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

    def test_install_download_argument_in_resulting_command(self):
        pkg = "pep8"
        download = "/tmp/foo"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, download=download)
            expected = [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--download",
                download,
                pkg,
            ]
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

    def test_install_no_download_argument_in_resulting_command(self):
        pkg = "pep8"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, no_download=True)
            expected = [sys.executable, "-m", "pip", "install", "--no-download", pkg]
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

    def test_install_download_cache_dir_arguments_in_resulting_command(self):
        pkg = "pep8"
        cache_dir_arg_mapping = {
            "1.5.6": "--download-cache",
            "6.0": "--cache-dir",
        }
        download_cache = "/tmp/foo"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})

        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            for pip_version, cmd_arg in cache_dir_arg_mapping.items():
                with patch(
                    "salt.modules.pip.version", MagicMock(return_value=pip_version)
                ):
                    # test `download_cache` kwarg
                    pip.install(pkg, download_cache="/tmp/foo")
                    expected = [
                        sys.executable,
                        "-m",
                        "pip",
                        "install",
                        cmd_arg,
                        download_cache,
                        pkg,
                    ]
                    mock.assert_called_with(
                        expected,
                        saltenv="base",
                        runas=None,
                        use_vt=False,
                        python_shell=False,
                    )

                    # test `cache_dir` kwarg
                    pip.install(pkg, cache_dir="/tmp/foo")
                    mock.assert_called_with(
                        expected,
                        saltenv="base",
                        runas=None,
                        use_vt=False,
                        python_shell=False,
                    )

    def test_install_source_argument_in_resulting_command(self):
        pkg = "pep8"
        source = "/tmp/foo"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, source=source)
            expected = [sys.executable, "-m", "pip", "install", "--source", source, pkg]
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

    def test_install_exists_action_argument_in_resulting_command(self):
        pkg = "pep8"
        for action in ("s", "i", "w", "b"):
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                pip.install(pkg, exists_action=action)
                expected = [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "--exists-action",
                    action,
                    pkg,
                ]
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

        # Test for invalid action
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            self.assertRaises(
                CommandExecutionError, pip.install, pkg, exists_action="d"
            )

    def test_install_install_options_argument_in_resulting_command(self):
        install_options = ["--exec-prefix=/foo/bar", "--install-scripts=/foo/bar/bin"]
        pkg = "pep8"

        expected = [sys.executable, "-m", "pip", "install"]
        for item in install_options:
            expected.extend(["--install-option", item])
        expected.append(pkg)

        # Passing options as a list
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, install_options=install_options)
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

        # Passing mirrors as a comma separated list
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, install_options=",".join(install_options))
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

        # Passing mirrors as a single string entry
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, install_options=install_options[0])
            expected = [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--install-option",
                install_options[0],
                pkg,
            ]
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

    def test_install_global_options_argument_in_resulting_command(self):
        global_options = ["--quiet", "--no-user-cfg"]
        pkg = "pep8"

        expected = [sys.executable, "-m", "pip", "install"]
        for item in global_options:
            expected.extend(["--global-option", item])
        expected.append(pkg)

        # Passing options as a list
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, global_options=global_options)
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

        # Passing mirrors as a comma separated list
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, global_options=",".join(global_options))
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

        # Passing mirrors as a single string entry
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, global_options=global_options[0])
            expected = [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--global-option",
                global_options[0],
                pkg,
            ]
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

    def test_install_upgrade_argument_in_resulting_command(self):
        pkg = "pep8"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, upgrade=True)
            expected = [sys.executable, "-m", "pip", "install", "--upgrade", pkg]
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

    def test_install_force_reinstall_argument_in_resulting_command(self):
        pkg = "pep8"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, force_reinstall=True)
            expected = [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--force-reinstall",
                pkg,
            ]
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

    def test_install_ignore_installed_argument_in_resulting_command(self):
        pkg = "pep8"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, ignore_installed=True)
            expected = [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--ignore-installed",
                pkg,
            ]
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

    def test_install_no_deps_argument_in_resulting_command(self):
        pkg = "pep8"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, no_deps=True)
            expected = [sys.executable, "-m", "pip", "install", "--no-deps", pkg]
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

    def test_install_no_install_argument_in_resulting_command(self):
        pkg = "pep8"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, no_install=True)
            expected = [sys.executable, "-m", "pip", "install", "--no-install", pkg]
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

    def test_install_proxy_argument_in_resulting_command(self):
        pkg = "pep8"
        proxy = "salt-user:salt-passwd@salt-proxy:3128"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, proxy=proxy)
            expected = [sys.executable, "-m", "pip", "install", "--proxy", proxy, pkg]
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

    def test_install_proxy_false_argument_in_resulting_command(self):
        """
        Checking that there is no proxy set if proxy arg is set to False
        even if the global proxy is set.
        """
        pkg = "pep8"
        proxy = False
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        config_mock = {
            "proxy_host": "salt-proxy",
            "proxy_port": "3128",
            "proxy_username": "salt-user",
            "proxy_password": "salt-passwd",
        }
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch.dict(pip.__opts__, config_mock):
                pip.install(pkg, proxy=proxy)
                expected = [sys.executable, "-m", "pip", "install", pkg]
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

    def test_install_global_proxy_in_resulting_command(self):
        """
        Checking that there is proxy set if global proxy is set.
        """
        pkg = "pep8"
        proxy = "http://salt-user:salt-passwd@salt-proxy:3128"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        config_mock = {
            "proxy_host": "salt-proxy",
            "proxy_port": "3128",
            "proxy_username": "salt-user",
            "proxy_password": "salt-passwd",
        }
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch.dict(pip.__opts__, config_mock):
                pip.install(pkg)
                expected = [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "--proxy",
                    proxy,
                    pkg,
                ]
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

    def test_install_multiple_requirements_arguments_in_resulting_command(self):
        with patch(
            "salt.modules.pip._get_cached_requirements"
        ) as get_cached_requirements:
            cached_reqs = ["my_cached_reqs-1", "my_cached_reqs-2"]
            get_cached_requirements.side_effect = cached_reqs
            requirements = ["salt://requirements-1.txt", "salt://requirements-2.txt"]

            expected = [sys.executable, "-m", "pip", "install"]
            for item in cached_reqs:
                expected.extend(["--requirement", item])

            # Passing option as a list
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                pip.install(requirements=requirements)
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

            # Passing option as a comma separated list
            get_cached_requirements.side_effect = cached_reqs
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                pip.install(requirements=",".join(requirements))
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

            # Passing option as a single string entry
            get_cached_requirements.side_effect = [cached_reqs[0]]
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                pip.install(requirements=requirements[0])
                expected = [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "--requirement",
                    cached_reqs[0],
                ]
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

    def test_install_extra_args_arguments_in_resulting_command(self):
        pkg = "pep8"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(
                pkg, extra_args=[{"--latest-pip-kwarg": "param"}, "--latest-pip-arg"]
            )
            expected = [
                sys.executable,
                "-m",
                "pip",
                "install",
                pkg,
                "--latest-pip-kwarg",
                "param",
                "--latest-pip-arg",
            ]
            mock.assert_called_with(
                expected, saltenv="base", runas=None, use_vt=False, python_shell=False,
            )

    def test_install_extra_args_arguments_recursion_error(self):
        pkg = "pep8"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):

            self.assertRaises(
                TypeError,
                lambda: pip.install(
                    pkg, extra_args=[{"--latest-pip-kwarg": ["param1", "param2"]}]
                ),
            )

            self.assertRaises(
                TypeError,
                lambda: pip.install(
                    pkg, extra_args=[{"--latest-pip-kwarg": [{"--too-deep": dict()}]}]
                ),
            )

    def test_uninstall_multiple_requirements_arguments_in_resulting_command(self):
        with patch(
            "salt.modules.pip._get_cached_requirements"
        ) as get_cached_requirements:
            cached_reqs = ["my_cached_reqs-1", "my_cached_reqs-2"]
            get_cached_requirements.side_effect = cached_reqs
            requirements = ["salt://requirements-1.txt", "salt://requirements-2.txt"]

            expected = [sys.executable, "-m", "pip", "uninstall", "-y"]
            for item in cached_reqs:
                expected.extend(["--requirement", item])

            # Passing option as a list
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                pip.uninstall(requirements=requirements)
                mock.assert_called_with(
                    expected,
                    cwd=None,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

            # Passing option as a comma separated list
            get_cached_requirements.side_effect = cached_reqs
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                pip.uninstall(requirements=",".join(requirements))
                mock.assert_called_with(
                    expected,
                    cwd=None,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

            # Passing option as a single string entry
            get_cached_requirements.side_effect = [cached_reqs[0]]
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                pip.uninstall(requirements=requirements[0])
                expected = [
                    sys.executable,
                    "-m",
                    "pip",
                    "uninstall",
                    "-y",
                    "--requirement",
                    cached_reqs[0],
                ]
                mock.assert_called_with(
                    expected,
                    cwd=None,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

    def test_uninstall_global_proxy_in_resulting_command(self):
        """
        Checking that there is proxy set if global proxy is set.
        """
        pkg = "pep8"
        proxy = "http://salt-user:salt-passwd@salt-proxy:3128"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        config_mock = {
            "proxy_host": "salt-proxy",
            "proxy_port": "3128",
            "proxy_username": "salt-user",
            "proxy_password": "salt-passwd",
        }
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch.dict(pip.__opts__, config_mock):
                pip.uninstall(pkg)
                expected = [
                    sys.executable,
                    "-m",
                    "pip",
                    "uninstall",
                    "-y",
                    "--proxy",
                    proxy,
                    pkg,
                ]
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    cwd=None,
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

    def test_uninstall_proxy_false_argument_in_resulting_command(self):
        """
        Checking that there is no proxy set if proxy arg is set to False
        even if the global proxy is set.
        """
        pkg = "pep8"
        proxy = False
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        config_mock = {
            "proxy_host": "salt-proxy",
            "proxy_port": "3128",
            "proxy_username": "salt-user",
            "proxy_password": "salt-passwd",
        }
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch.dict(pip.__opts__, config_mock):
                pip.uninstall(pkg, proxy=proxy)
                expected = [sys.executable, "-m", "pip", "uninstall", "-y", pkg]
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    cwd=None,
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

    def test_uninstall_log_argument_in_resulting_command(self):
        pkg = "pep8"
        log_path = "/tmp/pip-install.log"

        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.uninstall(pkg, log=log_path)
            expected = [
                sys.executable,
                "-m",
                "pip",
                "uninstall",
                "-y",
                "--log",
                log_path,
                pkg,
            ]
            mock.assert_called_with(
                expected,
                saltenv="base",
                cwd=None,
                runas=None,
                use_vt=False,
                python_shell=False,
            )

        # Let's fake a non-writable log file
        with patch("os.path") as mock_path:
            mock_path.exists.side_effect = IOError("Fooo!")
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                self.assertRaises(IOError, pip.uninstall, pkg, log=log_path)

    def test_uninstall_timeout_argument_in_resulting_command(self):
        pkg = "pep8"
        expected = [sys.executable, "-m", "pip", "uninstall", "-y", "--timeout"]
        # Passing an int
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.uninstall(pkg, timeout=10)
            mock.assert_called_with(
                expected + [10, pkg],
                cwd=None,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

        # Passing an int as a string
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.uninstall(pkg, timeout="10")
            mock.assert_called_with(
                expected + ["10", pkg],
                cwd=None,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

        # Passing a non-int to timeout
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            self.assertRaises(ValueError, pip.uninstall, pkg, timeout="a")

    def test_freeze_command(self):
        expected = [sys.executable, "-m", "pip", "freeze"]
        eggs = [
            "M2Crypto==0.21.1",
            "-e git+git@github.com:s0undt3ch/salt-testing.git@9ed81aa2f918d59d3706e56b18f0782d1ea43bf8#egg=SaltTesting-dev",
            "bbfreeze==1.1.0",
            "bbfreeze-loader==1.1.0",
            "pycrypto==2.6",
        ]
        mock = MagicMock(return_value={"retcode": 0, "stdout": "\n".join(eggs)})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch("salt.modules.pip.version", MagicMock(return_value="6.1.1")):
                ret = pip.freeze()
                mock.assert_called_with(
                    expected, cwd=None, runas=None, use_vt=False, python_shell=False,
                )
                self.assertEqual(ret, eggs)

        mock = MagicMock(return_value={"retcode": 0, "stdout": "\n".join(eggs)})
        # Passing env_vars passes them to underlying command?
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch("salt.modules.pip.version", MagicMock(return_value="6.1.1")):
                ret = pip.freeze(env_vars={"foo": "bar"})
                mock.assert_called_with(
                    expected,
                    cwd=None,
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                    env={"foo": "bar"},
                )
                self.assertEqual(ret, eggs)

        # Non zero returncode raises exception?
        mock = MagicMock(return_value={"retcode": 1, "stderr": "CABOOOOMMM!"})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch("salt.modules.pip.version", MagicMock(return_value="6.1.1")):
                self.assertRaises(
                    CommandExecutionError, pip.freeze,
                )

    def test_freeze_command_with_all(self):
        eggs = [
            "M2Crypto==0.21.1",
            "-e git+git@github.com:s0undt3ch/salt-testing.git@9ed81aa2f918d59d3706e56b18f0782d1ea43bf8#egg=SaltTesting-dev",
            "bbfreeze==1.1.0",
            "bbfreeze-loader==1.1.0",
            "pip==0.9.1",
            "pycrypto==2.6",
            "setuptools==20.10.1",
        ]
        mock = MagicMock(return_value={"retcode": 0, "stdout": "\n".join(eggs)})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch("salt.modules.pip.version", MagicMock(return_value="9.0.1")):
                ret = pip.freeze()
                expected = [sys.executable, "-m", "pip", "freeze", "--all"]
                mock.assert_called_with(
                    expected, cwd=None, runas=None, use_vt=False, python_shell=False,
                )
                self.assertEqual(ret, eggs)

        # Non zero returncode raises exception?
        mock = MagicMock(return_value={"retcode": 1, "stderr": "CABOOOOMMM!"})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch("salt.modules.pip.version", MagicMock(return_value="9.0.1")):
                self.assertRaises(
                    CommandExecutionError, pip.freeze,
                )

    def test_list_command(self):
        eggs = [
            "M2Crypto==0.21.1",
            "-e git+git@github.com:s0undt3ch/salt-testing.git@9ed81aa2f918d59d3706e56b18f0782d1ea43bf8#egg=SaltTesting-dev",
            "bbfreeze==1.1.0",
            "bbfreeze-loader==1.1.0",
            "pycrypto==2.6",
        ]
        mock_version = "6.1.1"
        mock = MagicMock(return_value={"retcode": 0, "stdout": "\n".join(eggs)})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch(
                "salt.modules.pip.version", MagicMock(return_value=mock_version)
            ):
                ret = pip.list_()
                expected = [sys.executable, "-m", "pip", "freeze"]
                mock.assert_called_with(
                    expected, cwd=None, runas=None, python_shell=False, use_vt=False,
                )
                self.assertEqual(
                    ret,
                    {
                        "SaltTesting-dev": "git+git@github.com:s0undt3ch/salt-testing.git@9ed81aa2f918d59d3706e56b18f0782d1ea43bf8",
                        "M2Crypto": "0.21.1",
                        "bbfreeze-loader": "1.1.0",
                        "bbfreeze": "1.1.0",
                        "pip": mock_version,
                        "pycrypto": "2.6",
                    },
                )

        # Non zero returncode raises exception?
        mock = MagicMock(return_value={"retcode": 1, "stderr": "CABOOOOMMM!"})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch("salt.modules.pip.version", MagicMock(return_value="6.1.1")):
                self.assertRaises(
                    CommandExecutionError, pip.list_,
                )

    def test_list_command_with_all(self):
        eggs = [
            "M2Crypto==0.21.1",
            "-e git+git@github.com:s0undt3ch/salt-testing.git@9ed81aa2f918d59d3706e56b18f0782d1ea43bf8#egg=SaltTesting-dev",
            "bbfreeze==1.1.0",
            "bbfreeze-loader==1.1.0",
            "pip==9.0.1",
            "pycrypto==2.6",
            "setuptools==20.10.1",
        ]
        # N.B.: this is deliberately different from the "output" of pip freeze.
        # This is to demonstrate that the version reported comes from freeze
        # instead of from the pip.version function.
        mock_version = "9.0.0"
        mock = MagicMock(return_value={"retcode": 0, "stdout": "\n".join(eggs)})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch(
                "salt.modules.pip.version", MagicMock(return_value=mock_version)
            ):
                ret = pip.list_()
                expected = [sys.executable, "-m", "pip", "freeze", "--all"]
                mock.assert_called_with(
                    expected, cwd=None, runas=None, python_shell=False, use_vt=False,
                )
                self.assertEqual(
                    ret,
                    {
                        "SaltTesting-dev": "git+git@github.com:s0undt3ch/salt-testing.git@9ed81aa2f918d59d3706e56b18f0782d1ea43bf8",
                        "M2Crypto": "0.21.1",
                        "bbfreeze-loader": "1.1.0",
                        "bbfreeze": "1.1.0",
                        "pip": "9.0.1",
                        "pycrypto": "2.6",
                        "setuptools": "20.10.1",
                    },
                )

        # Non zero returncode raises exception?
        mock = MagicMock(return_value={"retcode": 1, "stderr": "CABOOOOMMM!"})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch("salt.modules.pip.version", MagicMock(return_value="6.1.1")):
                self.assertRaises(
                    CommandExecutionError, pip.list_,
                )

    def test_list_command_with_prefix(self):
        eggs = [
            "M2Crypto==0.21.1",
            "-e git+git@github.com:s0undt3ch/salt-testing.git@9ed81aa2f918d59d3706e56b18f0782d1ea43bf8#egg=SaltTesting-dev",
            "bbfreeze==1.1.0",
            "bbfreeze-loader==1.1.0",
            "pycrypto==2.6",
        ]
        mock = MagicMock(return_value={"retcode": 0, "stdout": "\n".join(eggs)})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch("salt.modules.pip.version", MagicMock(return_value="6.1.1")):
                ret = pip.list_(prefix="bb")
                expected = [sys.executable, "-m", "pip", "freeze"]
                mock.assert_called_with(
                    expected, cwd=None, runas=None, python_shell=False, use_vt=False,
                )
                self.assertEqual(ret, {"bbfreeze-loader": "1.1.0", "bbfreeze": "1.1.0"})

    def test_list_upgrades_legacy(self):
        eggs = [
            "apache-libcloud (Current: 1.1.0 Latest: 2.2.1 [wheel])",
            "appdirs (Current: 1.4.1 Latest: 1.4.3 [wheel])",
            "awscli (Current: 1.11.63 Latest: 1.12.1 [sdist])",
        ]
        mock = MagicMock(return_value={"retcode": 0, "stdout": "\n".join(eggs)})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch("salt.modules.pip.version", MagicMock(return_value="6.1.1")):
                ret = pip.list_upgrades()
                mock.assert_called_with(
                    [sys.executable, "-m", "pip", "list", "--outdated"],
                    cwd=None,
                    runas=None,
                )
                self.assertEqual(
                    ret,
                    {
                        "apache-libcloud": "2.2.1 [wheel]",
                        "appdirs": "1.4.3 [wheel]",
                        "awscli": "1.12.1 [sdist]",
                    },
                )

    def test_list_upgrades_gt9(self):
        eggs = """[{"latest_filetype": "wheel", "version": "1.1.0", "name": "apache-libcloud", "latest_version": "2.2.1"},
                {"latest_filetype": "wheel", "version": "1.4.1", "name": "appdirs", "latest_version": "1.4.3"},
                {"latest_filetype": "sdist", "version": "1.11.63", "name": "awscli", "latest_version": "1.12.1"}
                ]"""
        mock = MagicMock(return_value={"retcode": 0, "stdout": "{0}".format(eggs)})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch("salt.modules.pip.version", MagicMock(return_value="9.1.1")):
                ret = pip.list_upgrades()
                mock.assert_called_with(
                    [
                        sys.executable,
                        "-m",
                        "pip",
                        "list",
                        "--outdated",
                        "--format=json",
                    ],
                    cwd=None,
                    runas=None,
                )
                self.assertEqual(
                    ret,
                    {
                        "apache-libcloud": "2.2.1 [wheel]",
                        "appdirs": "1.4.3 [wheel]",
                        "awscli": "1.12.1 [sdist]",
                    },
                )

    def test_is_installed_true(self):
        eggs = [
            "M2Crypto==0.21.1",
            "-e git+git@github.com:s0undt3ch/salt-testing.git@9ed81aa2f918d59d3706e56b18f0782d1ea43bf8#egg=SaltTesting-dev",
            "bbfreeze==1.1.0",
            "bbfreeze-loader==1.1.0",
            "pycrypto==2.6",
        ]
        mock = MagicMock(return_value={"retcode": 0, "stdout": "\n".join(eggs)})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch("salt.modules.pip.version", MagicMock(return_value="6.1.1")):
                ret = pip.is_installed(pkgname="bbfreeze")
                mock.assert_called_with(
                    [sys.executable, "-m", "pip", "freeze"],
                    cwd=None,
                    runas=None,
                    python_shell=False,
                    use_vt=False,
                )
                self.assertTrue(ret)

    def test_is_installed_false(self):
        eggs = [
            "M2Crypto==0.21.1",
            "-e git+git@github.com:s0undt3ch/salt-testing.git@9ed81aa2f918d59d3706e56b18f0782d1ea43bf8#egg=SaltTesting-dev",
            "bbfreeze==1.1.0",
            "bbfreeze-loader==1.1.0",
            "pycrypto==2.6",
        ]
        mock = MagicMock(return_value={"retcode": 0, "stdout": "\n".join(eggs)})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch("salt.modules.pip.version", MagicMock(return_value="6.1.1")):
                ret = pip.is_installed(pkgname="notexist")
                mock.assert_called_with(
                    [sys.executable, "-m", "pip", "freeze"],
                    cwd=None,
                    runas=None,
                    python_shell=False,
                    use_vt=False,
                )
                self.assertFalse(ret)

    def test_install_pre_argument_in_resulting_command(self):
        pkg = "pep8"
        # Lower than 1.4 versions don't end up with `--pre` in the resulting output
        mock = MagicMock(
            side_effect=[
                {"retcode": 0, "stdout": "pip 1.2.0 /path/to/site-packages/pip"},
                {"retcode": 0, "stdout": ""},
            ]
        )
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch("salt.modules.pip.version", MagicMock(return_value="1.3")):
                pip.install(pkg, pre_releases=True)
                expected = [sys.executable, "-m", "pip", "install", pkg]
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

        mock_run = MagicMock(return_value="pip 1.4.1 /path/to/site-packages/pip")
        mock_run_all = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(
            pip.__salt__, {"cmd.run_stdout": mock_run, "cmd.run_all": mock_run_all}
        ):
            with patch(
                "salt.modules.pip._get_pip_bin", MagicMock(return_value=["pip"])
            ):
                pip.install(pkg, pre_releases=True)
                expected = ["pip", "install", "--pre", pkg]
                mock_run_all.assert_called_with(
                    expected,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )



# -*- coding: utf-8 -*-
"""
    :codeauthor: Ted Strzalkowski (tedski@gmail.com)


    tests.unit.modules.mdadm_test
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import salt libs
import salt.modules.mdadm_raid as mdadm

# Import Salt Testing libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class MdadmTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {mdadm: {}}

    def test_create(self):
        mock = MagicMock(return_value="salt")
        with patch.dict(mdadm.__salt__, {"cmd.run": mock}), patch(
            "salt.utils.path.which", lambda exe: exe
        ):
            ret = mdadm.create(
                "/dev/md0",
                5,
                devices=["/dev/sdb1", "/dev/sdc1", "/dev/sdd1"],
                test_mode=False,
                force=True,
                chunk=256,
            )
            self.assertEqual("salt", ret)

            self.assert_called_once(mock)

            args, kwargs = mock.call_args
            # expected cmd is
            # mdadm -C /dev/md0 -R -v --chunk 256 --force -l 5 -e default -n 3 /dev/sdb1 /dev/sdc1 /dev/sdd1
            # where args between -v and -l could be in any order
            self.assertEqual(len(args), 1)
            self.assertEqual(len(args[0]), 17)
            self.assertEqual(
                args[0][:7], ["mdadm", "-C", "/dev/md0", "-R", "-v", "-l", "5"]
            )
            self.assertEqual(
                args[0][10:],
                ["-e", "default", "-n", "3", "/dev/sdb1", "/dev/sdc1", "/dev/sdd1"],
            )
            self.assertEqual(
                sorted(args[0][7:10]), sorted(["--chunk", "256", "--force"])
            )
            self.assertEqual(kwargs, {"python_shell": False})

    def test_create_metadata(self):
        mock = MagicMock(return_value="salt")
        with patch.dict(mdadm.__salt__, {"cmd.run": mock}), patch(
            "salt.utils.path.which", lambda exe: exe
        ):
            ret = mdadm.create(
                "/dev/md0",
                5,
                devices=["/dev/sdb1", "/dev/sdc1", "/dev/sdd1"],
                metadata=0.9,
                test_mode=False,
                force=True,
                chunk=256,
            )
            self.assertEqual("salt", ret)

            self.assert_called_once(mock)

            args, kwargs = mock.call_args
            self.assertEqual(
                args[0][:7], ["mdadm", "-C", "/dev/md0", "-R", "-v", "-l", "5"]
            )
            self.assertEqual(
                args[0][10:],
                ["-e", "0.9", "-n", "3", "/dev/sdb1", "/dev/sdc1", "/dev/sdd1"],
            )
            self.assertEqual(
                sorted(args[0][7:10]), sorted(["--chunk", "256", "--force"])
            )
            self.assertEqual(kwargs, {"python_shell": False})

    def test_create_test_mode(self):
        mock = MagicMock()
        with patch.dict(mdadm.__salt__, {"cmd.run": mock}):
            ret = mdadm.create(
                "/dev/md0",
                5,
                devices=["/dev/sdb1", "/dev/sdc1", "/dev/sdd1"],
                force=True,
                chunk=256,
                test_mode=True,
            )
            self.assertEqual(
                sorted(
                    "mdadm -C /dev/md0 -R -v --chunk 256 "
                    "--force -l 5 -e default -n 3 "
                    "/dev/sdb1 /dev/sdc1 /dev/sdd1".split()
                ),
                sorted(ret.split()),
            )
            assert not mock.called, "test mode failed, cmd.run called"

    def test_examine(self):
        """
        Test for mdadm_raid.examine
        """
        mock = MagicMock(
            return_value="ARRAY /dev/md/pool metadata=1.2 UUID=567da122:fb8e445e:55b853e0:81bd0a3e name=positron:pool"
        )
        with patch.dict(mdadm.__salt__, {"cmd.run_stdout": mock}):
            self.assertEqual(
                mdadm.examine("/dev/md0"),
                {
                    "ARRAY /dev/md/pool metadata": "1.2 UUID=567da122:fb8e445e:55b853e0:81bd0a3e name=positron:pool"
                },
            )
            mock.assert_called_with(
                "mdadm -Y -E /dev/md0", ignore_retcode=False, python_shell=False
            )

    def test_examine_quiet(self):
        """
        Test for mdadm_raid.examine
        """
        mock = MagicMock(return_value="")
        with patch.dict(mdadm.__salt__, {"cmd.run_stdout": mock}):
            self.assertEqual(mdadm.examine("/dev/md0", quiet=True), {})
            mock.assert_called_with(
                "mdadm -Y -E /dev/md0", ignore_retcode=True, python_shell=False
            )



# -*- coding: utf-8 -*-
"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""
# Import python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.modules.mac_user as mac_user
from salt.exceptions import CommandExecutionError, SaltInvocationError

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase, skipIf

HAS_PWD = True
try:
    import pwd
except ImportError:
    HAS_PWD = False


@skipIf(not HAS_PWD, "Missing required library 'pwd'")
class MacUserTestCase(TestCase, LoaderModuleMockMixin):
    """
    TestCase for the salt.modules.mac_user modules
    """

    def setup_loader_modules(self):
        return {mac_user: {}}

    if HAS_PWD:
        mock_pwall = [
            pwd.struct_passwd(
                (
                    "_amavisd",
                    "*",
                    83,
                    83,
                    "AMaViS Daemon",
                    "/var/virusmails",
                    "/usr/bin/false",
                )
            ),
            pwd.struct_passwd(
                (
                    "_appleevents",
                    "*",
                    55,
                    55,
                    "AppleEvents Daemon",
                    "/var/empty",
                    "/usr/bin/false",
                )
            ),
            pwd.struct_passwd(
                (
                    "_appowner",
                    "*",
                    87,
                    87,
                    "Application Owner",
                    "/var/empty",
                    "/usr/bin/false",
                )
            ),
        ]
    mock_info_ret = {
        "shell": "/bin/bash",
        "name": "test",
        "gid": 4376,
        "groups": ["TEST_GROUP"],
        "home": "/Users/foo",
        "fullname": "TEST USER",
        "uid": 4376,
    }

    @classmethod
    def tearDownClass(cls):
        for attrname in ("mock_pwall", "mock_info_ret"):
            delattr(cls, attrname)

    @skipIf(True, "Waiting on some clarifications from bug report #10594")
    def _test_flush_dscl_cache(self):
        # TODO: Implement tests after clarifications come in
        pass

    def test_dscl(self):
        """
        Tests the creation of a dscl node
        """
        mac_mock = MagicMock(
            return_value={"pid": 4948, "retcode": 0, "stderr": "", "stdout": ""}
        )
        with patch.dict(mac_user.__salt__, {"cmd.run_all": mac_mock}):
            with patch.dict(
                mac_user.__grains__,
                {
                    "kernel": "Darwin",
                    "osrelease": "10.9.1",
                    "osrelease_info": (10, 9, 1),
                },
            ):
                self.assertEqual(
                    mac_user._dscl(["username", "UniqueID", 501]),
                    {"pid": 4948, "retcode": 0, "stderr": "", "stdout": ""},
                )

    def test_first_avail_uid(self):
        """
        Tests the availability of the next uid
        """
        with patch("pwd.getpwall", MagicMock(return_value=self.mock_pwall)):
            self.assertEqual(mac_user._first_avail_uid(), 501)

    # 'add' function tests: 4
    # Only tested error handling
    # Full functionality tests covered in integration testing

    def test_add_user_exists(self):
        """
        Tests if the user exists or not
        """
        with patch(
            "salt.modules.mac_user.info", MagicMock(return_value=self.mock_info_ret)
        ):
            self.assertRaises(CommandExecutionError, mac_user.add, "test")

    def test_add_whitespace(self):
        """
        Tests if there is whitespace in the user name
        """
        with patch("salt.modules.mac_user.info", MagicMock(return_value={})):
            self.assertRaises(SaltInvocationError, mac_user.add, "foo bar")

    def test_add_uid_int(self):
        """
        Tests if the uid is an int
        """
        with patch("salt.modules.mac_user.info", MagicMock(return_value={})):
            self.assertRaises(SaltInvocationError, mac_user.add, "foo", "foo")

    def test_add_gid_int(self):
        """
        Tests if the gid is an int
        """
        with patch("salt.modules.mac_user.info", MagicMock(return_value={})):
            self.assertRaises(SaltInvocationError, mac_user.add, "foo", 20, "foo")

    # 'delete' function tests: 2
    # Only tested pure logic of function
    # Full functionality tests covered in integration testing

    def test_delete_whitespace(self):
        """
        Tests if there is whitespace in the user name
        """
        self.assertRaises(SaltInvocationError, mac_user.delete, "foo bar")

    def test_delete_user_exists(self):
        """
        Tests if the user exists or not
        """
        with patch("salt.modules.mac_user.info", MagicMock(return_value={})):
            self.assertTrue(mac_user.delete("foo"))

    def test_getent(self):
        """
        Tests the list of information for all users
        """
        with patch("pwd.getpwall", MagicMock(return_value=self.mock_pwall)), patch(
            "salt.modules.mac_user.list_groups", MagicMock(return_value=["TEST_GROUP"])
        ):
            ret = [
                {
                    "shell": "/usr/bin/false",
                    "name": "_amavisd",
                    "gid": 83,
                    "groups": ["TEST_GROUP"],
                    "home": "/var/virusmails",
                    "fullname": "AMaViS Daemon",
                    "uid": 83,
                },
                {
                    "shell": "/usr/bin/false",
                    "name": "_appleevents",
                    "gid": 55,
                    "groups": ["TEST_GROUP"],
                    "home": "/var/empty",
                    "fullname": "AppleEvents Daemon",
                    "uid": 55,
                },
                {
                    "shell": "/usr/bin/false",
                    "name": "_appowner",
                    "gid": 87,
                    "groups": ["TEST_GROUP"],
                    "home": "/var/empty",
                    "fullname": "Application Owner",
                    "uid": 87,
                },
            ]
            self.assertEqual(mac_user.getent(), ret)

    # 'chuid' function tests: 3
    # Only tested pure logic of function
    # Full functionality tests covered in integration testing

    def test_chuid_int(self):
        """
        Tests if the uid is an int
        """
        self.assertRaises(SaltInvocationError, mac_user.chuid, "foo", "foo")

    def test_chuid_user_exists(self):
        """
        Tests if the user exists or not
        """
        with patch("salt.modules.mac_user.info", MagicMock(return_value={})):
            self.assertRaises(CommandExecutionError, mac_user.chuid, "foo", 4376)

    def test_chuid_same_uid(self):
        """
        Tests if the user's uid is the same as as the argument
        """
        with patch(
            "salt.modules.mac_user.info", MagicMock(return_value=self.mock_info_ret)
        ):
            self.assertTrue(mac_user.chuid("foo", 4376))

    # 'chgid' function tests: 3
    # Only tested pure logic of function
    # Full functionality tests covered in integration testing

    def test_chgid_int(self):
        """
        Tests if the gid is an int
        """
        self.assertRaises(SaltInvocationError, mac_user.chgid, "foo", "foo")

    def test_chgid_user_exists(self):
        """
        Tests if the user exists or not
        """
        with patch("salt.modules.mac_user.info", MagicMock(return_value={})):
            self.assertRaises(CommandExecutionError, mac_user.chgid, "foo", 4376)

    def test_chgid_same_gid(self):
        """
        Tests if the user's gid is the same as as the argument
        """
        with patch(
            "salt.modules.mac_user.info", MagicMock(return_value=self.mock_info_ret)
        ):
            self.assertTrue(mac_user.chgid("foo", 4376))

    # 'chshell' function tests: 2
    # Only tested pure logic of function
    # Full functionality tests covered in integration testing

    def test_chshell_user_exists(self):
        """
        Tests if the user exists or not
        """
        with patch("salt.modules.mac_user.info", MagicMock(return_value={})):
            self.assertRaises(
                CommandExecutionError, mac_user.chshell, "foo", "/bin/bash"
            )

    def test_chshell_same_shell(self):
        """
        Tests if the user's shell is the same as the argument
        """
        with patch(
            "salt.modules.mac_user.info", MagicMock(return_value=self.mock_info_ret)
        ):
            self.assertTrue(mac_user.chshell("foo", "/bin/bash"))

    # 'chhome' function tests: 2
    # Only tested pure logic of function
    # Full functionality tests covered in integration testing

    def test_chhome_user_exists(self):
        """
        Test if the user exists or not
        """
        with patch("salt.modules.mac_user.info", MagicMock(return_value={})):
            self.assertRaises(
                CommandExecutionError, mac_user.chhome, "foo", "/Users/foo"
            )

    def test_chhome_same_home(self):
        """
        Tests if the user's home is the same as the argument
        """
        with patch(
            "salt.modules.mac_user.info", MagicMock(return_value=self.mock_info_ret)
        ):
            self.assertTrue(mac_user.chhome("foo", "/Users/foo"))

    # 'chfullname' function tests: 2
    # Only tested pure logic of function
    # Full functionality tests covered in integration testing

    def test_chfullname_user_exists(self):
        """
        Tests if the user exists or not
        """
        with patch("salt.modules.mac_user.info", MagicMock(return_value={})):
            self.assertRaises(
                CommandExecutionError, mac_user.chfullname, "test", "TEST USER"
            )

    def test_chfullname_same_name(self):
        """
        Tests if the user's full name is the same as the argument
        """
        with patch(
            "salt.modules.mac_user.info", MagicMock(return_value=self.mock_info_ret)
        ):
            self.assertTrue(mac_user.chfullname("test", "TEST USER"))

    # 'chgroups' function tests: 3
    # Only tested pure logic of function
    # Full functionality tests covered in integration testing

    def test_chgroups_user_exists(self):
        """
        Tests if the user exists or not
        """
        with patch("salt.modules.mac_user.info", MagicMock(return_value={})):
            self.assertRaises(
                CommandExecutionError, mac_user.chgroups, "foo", "wheel,root"
            )

    def test_chgroups_bad_groups(self):
        """
        Test if there is white space in groups argument
        """
        with patch(
            "salt.modules.mac_user.info", MagicMock(return_value=self.mock_info_ret)
        ):
            self.assertRaises(
                SaltInvocationError, mac_user.chgroups, "test", "bad group"
            )

    def test_chgroups_same_desired(self):
        """
        Tests if the user's list of groups is the same as the arguments
        """
        mock_primary = MagicMock(return_value="wheel")
        with patch.dict(mac_user.__salt__, {"file.gid_to_group": mock_primary}), patch(
            "salt.modules.mac_user.info", MagicMock(return_value=self.mock_info_ret)
        ), patch(
            "salt.modules.mac_user.list_groups",
            MagicMock(return_value=("wheel", "root")),
        ):
            self.assertTrue(mac_user.chgroups("test", "wheel,root"))

    def test_info(self):
        """
        Tests the return of user information
        """
        mock_pwnam = pwd.struct_passwd(
            ("test", "*", 0, 0, "TEST USER", "/var/test", "/bin/bash")
        )
        ret = {
            "shell": "/bin/bash",
            "name": "test",
            "gid": 0,
            "groups": ["_TEST_GROUP"],
            "home": "/var/test",
            "fullname": "TEST USER",
            "uid": 0,
        }
        with patch("pwd.getpwnam", MagicMock(return_value=mock_pwnam)), patch(
            "salt.modules.mac_user.list_groups", MagicMock(return_value=["_TEST_GROUP"])
        ):
            self.assertEqual(mac_user.info("root"), ret)

    def test_format_info(self):
        """
        Tests the formatting of returned user information
        """
        data = pwd.struct_passwd(
            (
                "_TEST_GROUP",
                "*",
                83,
                83,
                "AMaViS Daemon",
                "/var/virusmails",
                "/usr/bin/false",
            )
        )
        ret = {
            "shell": "/usr/bin/false",
            "name": "_TEST_GROUP",
            "gid": 83,
            "groups": ["_TEST_GROUP"],
            "home": "/var/virusmails",
            "fullname": "AMaViS Daemon",
            "uid": 83,
        }
        with patch(
            "salt.modules.mac_user.list_groups", MagicMock(return_value=["_TEST_GROUP"])
        ):
            self.assertEqual(mac_user._format_info(data), ret)

    def test_list_users(self):
        """
        Tests the list of all users
        """
        expected = ["spongebob", "patrick", "squidward"]
        mock_run = MagicMock(
            return_value={
                "pid": 4948,
                "retcode": 0,
                "stderr": "",
                "stdout": "\n".join(expected),
            }
        )
        with patch.dict(
            mac_user.__grains__, {"osrelease_info": (10, 9, 1)}
        ), patch.dict(mac_user.__salt__, {"cmd.run_all": mock_run}):
            self.assertEqual(mac_user.list_users(), expected)



# -*- coding: utf-8 -*-
"""
ACME / Let's Encrypt module
===========================

.. versionadded: 2016.3

This module currently looks for certbot script in the $PATH as
- certbot,
- lestsencrypt,
- certbot-auto,
- letsencrypt-auto
eventually falls back to /opt/letsencrypt/letsencrypt-auto

.. note::

    Installation & configuration of the Let's Encrypt client can for example be done using
    https://github.com/saltstack-formulas/letsencrypt-formula

.. warning::

    Be sure to set at least accept-tos = True in cli.ini!

Most parameters will fall back to cli.ini defaults if None is given.

DNS plugins
-----------

This module currently supports the CloudFlare certbot DNS plugin.  The DNS
plugin credentials file needs to be passed in using the
``dns_plugin_credentials`` argument.

Make sure the appropriate certbot plugin for the wanted DNS provider is
installed before using this module.

"""
# Import python libs
from __future__ import absolute_import, print_function, unicode_literals

import datetime
import logging
import os

# Import salt libs
import salt.utils.path
from salt.exceptions import SaltInvocationError

log = logging.getLogger(__name__)

LEA = salt.utils.path.which_bin(
    [
        "certbot",
        "letsencrypt",
        "certbot-auto",
        "letsencrypt-auto",
        "/opt/letsencrypt/letsencrypt-auto",
    ]
)
LE_LIVE = "/etc/letsencrypt/live/"

if salt.utils.platform.is_freebsd():
    LE_LIVE = "/usr/local" + LE_LIVE


def __virtual__():
    """
    Only work when letsencrypt-auto is installed
    """
    return (
        LEA is not None,
        "The ACME execution module cannot be loaded: letsencrypt-auto not installed.",
    )


def _cert_file(name, cert_type):
    """
    Return expected path of a Let's Encrypt live cert
    """
    return os.path.join(LE_LIVE, name, "{0}.pem".format(cert_type))


def _expires(name):
    """
    Return the expiry date of a cert

    :rtype: datetime
    :return: Expiry date
    """
    cert_file = _cert_file(name, "cert")
    # Use the salt module if available
    if "tls.cert_info" in __salt__:
        expiry = __salt__["tls.cert_info"](cert_file).get("not_after", 0)
    # Cobble it together using the openssl binary
    else:
        openssl_cmd = "openssl x509 -in {0} -noout -enddate".format(cert_file)
        # No %e format on my Linux'es here
        strptime_sux_cmd = 'date --date="$({0} | cut -d= -f2)" +%s'.format(openssl_cmd)
        expiry = float(__salt__["cmd.shell"](strptime_sux_cmd, output_loglevel="quiet"))
        # expiry = datetime.datetime.strptime(expiry.split('=', 1)[-1], '%b %e %H:%M:%S %Y %Z')
    return datetime.datetime.fromtimestamp(expiry)


def _renew_by(name, window=None):
    """
    Date before a certificate should be renewed

    :param str name: Common Name of the certificate (DNS name of certificate)
    :param int window: days before expiry date to renew
    :rtype: datetime
    :return: First renewal date
    """
    expiry = _expires(name)
    if window is not None:
        expiry = expiry - datetime.timedelta(days=window)

    return expiry


def cert(
    name,
    aliases=None,
    email=None,
    webroot=None,
    test_cert=False,
    renew=None,
    keysize=None,
    server=None,
    owner="root",
    group="root",
    mode="0640",
    certname=None,
    preferred_challenges=None,
    tls_sni_01_port=None,
    tls_sni_01_address=None,
    http_01_port=None,
    http_01_address=None,
    dns_plugin=None,
    dns_plugin_credentials=None,
):
    """
    Obtain/renew a certificate from an ACME CA, probably Let's Encrypt.

    :param name: Common Name of the certificate (DNS name of certificate)
    :param aliases: subjectAltNames (Additional DNS names on certificate)
    :param email: e-mail address for interaction with ACME provider
    :param webroot: True or a full path to use to use webroot. Otherwise use standalone mode
    :param test_cert: Request a certificate from the Happy Hacker Fake CA (mutually
        exclusive with 'server')
    :param renew: True/'force' to force a renewal, or a window of renewal before
        expiry in days
    :param keysize: RSA key bits
    :param server: API endpoint to talk to
    :param owner: owner of the private key file
    :param group: group of the private key file
    :param mode: mode of the private key file
    :param certname: Name of the certificate to save
    :param preferred_challenges: A sorted, comma delimited list of the preferred
        challenge to use during authorization with the most preferred challenge
        listed first.
    :param tls_sni_01_port: Port used during tls-sni-01 challenge. This only affects
        the port Certbot listens on. A conforming ACME server will still attempt
        to connect on port 443.
    :param tls_sni_01_address: The address the server listens to during tls-sni-01
        challenge.
    :param http_01_port: Port used in the http-01 challenge. This only affects
        the port Certbot listens on. A conforming ACME server will still attempt
        to connect on port 80.
    :param https_01_address: The address the server listens to during http-01 challenge.
    :param dns_plugin: Name of a DNS plugin to use (currently only 'cloudflare'
        or 'digitalocean')
    :param dns_plugin_credentials: Path to the credentials file if required by
        the specified DNS plugin
    :param dns_plugin_propagate_seconds: Number of seconds to wait for DNS propogations
        before asking ACME servers to verify the DNS record. (default 10)
    :rtype: dict
    :return: Dictionary with 'result' True/False/None, 'comment' and certificate's
        expiry date ('not_after')

    CLI example:

    .. code-block:: bash

        salt 'gitlab.example.com' acme.cert dev.example.com "[gitlab.example.com]" test_cert=True \
        renew=14 webroot=/opt/gitlab/embedded/service/gitlab-rails/public
    """

    cmd = [LEA, "certonly", "--non-interactive", "--agree-tos"]

    supported_dns_plugins = ["cloudflare"]

    cert_file = _cert_file(name, "cert")
    if not __salt__["file.file_exists"](cert_file):
        log.debug("Certificate %s does not exist (yet)", cert_file)
        renew = False
    elif needs_renewal(name, renew):
        log.debug("Certificate %s will be renewed", cert_file)
        cmd.append("--renew-by-default")
        renew = True
    if server:
        cmd.append("--server {0}".format(server))

    if certname:
        cmd.append("--cert-name {0}".format(certname))

    if test_cert:
        if server:
            return {
                "result": False,
                "comment": "Use either server or test_cert, not both",
            }
        cmd.append("--test-cert")

    if webroot:
        cmd.append("--authenticator webroot")
        if webroot is not True:
            cmd.append("--webroot-path {0}".format(webroot))
    elif dns_plugin in supported_dns_plugins:
        if dns_plugin == "cloudflare":
            cmd.append("--dns-cloudflare")
            cmd.append(
                "--dns-cloudflare-credentials {0}".format(dns_plugin_credentials)
            )
        else:
            return {
                "result": False,
                "comment": "DNS plugin '{0}' is not supported".format(dns_plugin),
            }
    else:
        cmd.append("--authenticator standalone")

    if email:
        cmd.append("--email {0}".format(email))

    if keysize:
        cmd.append("--rsa-key-size {0}".format(keysize))

    cmd.append("--domains {0}".format(name))
    if aliases is not None:
        for dns in aliases:
            cmd.append("--domains {0}".format(dns))

    if preferred_challenges:
        cmd.append("--preferred-challenges {}".format(preferred_challenges))

    if tls_sni_01_port:
        cmd.append("--tls-sni-01-port {}".format(tls_sni_01_port))
    if tls_sni_01_address:
        cmd.append("--tls-sni-01-address {}".format(tls_sni_01_address))
    if http_01_port:
        cmd.append("--http-01-port {}".format(http_01_port))
    if http_01_address:
        cmd.append("--http-01-address {}".format(http_01_address))

    res = __salt__["cmd.run_all"](" ".join(cmd))

    if res["retcode"] != 0:
        if "expand" in res["stderr"]:
            cmd.append("--expand")
            res = __salt__["cmd.run_all"](" ".join(cmd))
            if res["retcode"] != 0:
                return {
                    "result": False,
                    "comment": (
                        "Certificate {0} renewal failed with:\n{1}"
                        "".format(name, res["stderr"])
                    ),
                }
        else:
            return {
                "result": False,
                "comment": (
                    "Certificate {0} renewal failed with:\n{1}"
                    "".format(name, res["stderr"])
                ),
            }

    if "no action taken" in res["stdout"]:
        comment = "Certificate {0} unchanged".format(cert_file)
        result = None
    elif renew:
        comment = "Certificate {0} renewed".format(name)
        result = True
    else:
        comment = "Certificate {0} obtained".format(name)
        result = True

    ret = {
        "comment": comment,
        "not_after": expires(name),
        "changes": {},
        "result": result,
    }
    ret, _ = __salt__["file.check_perms"](
        _cert_file(name, "privkey"), ret, owner, group, mode, follow_symlinks=True
    )

    return ret


def certs():
    """
    Return a list of active certificates

    CLI example:

    .. code-block:: bash

        salt 'vhost.example.com' acme.certs
    """
    return [
        item
        for item in __salt__["file.readdir"](LE_LIVE)[2:]
        if os.path.isdir(os.path.join(LE_LIVE, item))
    ]


def info(name):
    """
    Return information about a certificate

    :param str name: CommonName of certificate
    :rtype: dict
    :return: Dictionary with information about the certificate.
        If neither the ``tls`` nor the ``x509`` module can be used to determine
        the certificate information, the information will be retrieved as one
        big text block under the key ``text`` using the openssl cli.

    CLI example:

    .. code-block:: bash

        salt 'gitlab.example.com' acme.info dev.example.com
    """
    if not has(name):
        return {}
    cert_file = _cert_file(name, "cert")
    # Use the tls salt module if available
    if "tls.cert_info" in __salt__:
        cert_info = __salt__["tls.cert_info"](cert_file)
        # Strip out the extensions object contents;
        # these trip over our poor state output
        # and they serve no real purpose here anyway
        cert_info["extensions"] = list(cert_info["extensions"])
    elif "x509.read_certificate" in __salt__:
        cert_info = __salt__["x509.read_certificate"](cert_file)
    else:
        # Cobble it together using the openssl binary
        openssl_cmd = "openssl x509 -in {0} -noout -text".format(cert_file)
        cert_info = {"text": __salt__["cmd.run"](openssl_cmd, output_loglevel="quiet")}
    return cert_info


def expires(name):
    """
    The expiry date of a certificate in ISO format

    :param str name: CommonName of certificate
    :rtype: str
    :return: Expiry date in ISO format.

    CLI example:

    .. code-block:: bash

        salt 'gitlab.example.com' acme.expires dev.example.com
    """
    return _expires(name).isoformat()


def has(name):
    """
    Test if a certificate is in the Let's Encrypt Live directory

    :param str name: CommonName of certificate
    :rtype: bool

    Code example:

    .. code-block:: python

        if __salt__['acme.has']('dev.example.com'):
            log.info('That is one nice certificate you have there!')
    """
    return __salt__["file.file_exists"](_cert_file(name, "cert"))


def renew_by(name, window=None):
    """
    Date in ISO format when a certificate should first be renewed

    :param str name: CommonName of certificate
    :param int window: number of days before expiry when renewal should take place
    :rtype: str
    :return: Date of certificate renewal in ISO format.
    """
    return _renew_by(name, window).isoformat()


def needs_renewal(name, window=None):
    """
    Check if a certificate needs renewal

    :param str name: CommonName of certificate
    :param bool/str/int window: Window in days to renew earlier or True/force to just return True
    :rtype: bool
    :return: Whether or not the certificate needs to be renewed.

    Code example:

    .. code-block:: python

        if __salt__['acme.needs_renewal']('dev.example.com'):
            __salt__['acme.cert']('dev.example.com', **kwargs)
        else:
            log.info('Your certificate is still good')
    """
    if window:
        if str(window).lower() in ("force", "true"):
            return True
        if not (
            isinstance(window, int) or (hasattr(window, "isdigit") and window.isdigit())
        ):
            raise SaltInvocationError(
                'The argument "window", if provided, must be one of the following : '
                'True (boolean), "force" or "Force" (str) or a numerical value in days.'
            )
        window = int(window)

    return _renew_by(name, window) <= datetime.datetime.today()



# -*- coding: utf-8 -*-

from __future__ import absolute_import, print_function, unicode_literals

import re

import salt.utils.platform
import salt.utils.pycrypto
from salt.exceptions import SaltInvocationError
from tests.support.mock import patch
from tests.support.unit import TestCase, skipIf


class PycryptoTestCase(TestCase):
    """
    TestCase for salt.utils.pycrypto module
    """

    passwd = "test_password"
    expecteds = {
        "sha512": {
            "hashed": "$6$rounds=65601$goodsalt$lZFhiN5M8RTLd9WKDin50H4lF4F8HGMIdwvKs.nTG7f8F0Y4P447Zb9/E8SkUWjY.K10QT3NuHZNDgc/P/NjT1",
            "salt": "rounds=65601$goodsalt",
            "badsalt": "badsalt",
        },
        "sha256": {
            "hashed": "$5$rounds=53501$goodsalt$W.uoco0wMfGLDOlsbW52E6raFS1Nhj0McfUTj2vORt7",
            "salt": "rounds=53501$goodsalt",
            "badsalt": "badsalt",
        },
        "blowfish": {
            "hashed": "$2b$10$goodsaltgoodsaltgoodsObFfGrJwfV.13QddrZIh2w1ccESmvj8K",
            "salt": "10$goodsaltgoodsaltgoodsa",
            "badsalt": "badsaltbadsaltbadsaltb",
        },
        "md5": {
            "hashed": "$1$goodsalt$4XQMx4a4e1MpBB8xzz.TQ0",
            "salt": "goodsalt",
            "badsalt": "badsalt",
        },
        "crypt": {"hashed": "goVHulDpuGA7w", "salt": "go", "badsalt": "ba"},
    }
    invalid_salt = "thissaltistoolong" * 10

    @skipIf(not salt.utils.pycrypto.HAS_CRYPT, "crypt not available")
    def test_gen_hash_crypt(self):
        """
        Test gen_hash with crypt library
        """
        methods = salt.utils.pycrypto.methods

        for algorithm in methods:
            expected = self.expecteds[algorithm]
            ret = salt.utils.pycrypto.gen_hash(
                crypt_salt=expected["salt"], password=self.passwd, algorithm=algorithm,
            )
            self.assertEqual(ret, expected["hashed"])

            ret = salt.utils.pycrypto.gen_hash(
                crypt_salt=expected["badsalt"],
                password=self.passwd,
                algorithm=algorithm,
            )
            self.assertNotEqual(ret, expected["hashed"])

            ret = salt.utils.pycrypto.gen_hash(
                crypt_salt=None, password=self.passwd, algorithm=algorithm
            )
            self.assertNotEqual(ret, expected["hashed"])

        # Assert it works without arguments passed
        self.assertIsNotNone(salt.utils.pycrypto.gen_hash())

        # Assert it works without algorithm passed
        default_algorithm = salt.utils.pycrypto.crypt.methods[0].name.lower()
        expected = self.expecteds[default_algorithm]
        ret = salt.utils.pycrypto.gen_hash(
            crypt_salt=expected["salt"], password=self.passwd,
        )
        self.assertEqual(ret, expected["hashed"])

    @skipIf(not salt.utils.pycrypto.HAS_PASSLIB, "passlib not available")
    @patch("salt.utils.pycrypto.methods", {})
    @patch("salt.utils.pycrypto.HAS_CRYPT", False)
    def test_gen_hash_passlib(self):
        """
        Test gen_hash with passlib
        """
        methods = salt.utils.pycrypto.known_methods

        for algorithm in methods:
            expected = self.expecteds[algorithm]
            ret = salt.utils.pycrypto.gen_hash(
                crypt_salt=expected["salt"], password=self.passwd, algorithm=algorithm,
            )
            self.assertEqual(ret, expected["hashed"])

            ret = salt.utils.pycrypto.gen_hash(
                crypt_salt=expected["badsalt"],
                password=self.passwd,
                algorithm=algorithm,
            )
            self.assertNotEqual(ret, expected["hashed"])

            ret = salt.utils.pycrypto.gen_hash(
                crypt_salt=None, password=self.passwd, algorithm=algorithm
            )
            self.assertNotEqual(ret, expected["hashed"])

        # Assert it works without arguments passed
        self.assertIsNotNone(salt.utils.pycrypto.gen_hash())

        # Assert it works without algorithm passed
        default_algorithm = salt.utils.pycrypto.known_methods[0]
        expected = self.expecteds[default_algorithm]
        if default_algorithm in self.expecteds:
            ret = salt.utils.pycrypto.gen_hash(
                crypt_salt=expected["salt"], password=self.passwd
            )
            self.assertEqual(ret, expected["hashed"])

    @patch("salt.utils.pycrypto.HAS_CRYPT", False)
    @patch("salt.utils.pycrypto.HAS_PASSLIB", False)
    def test_gen_hash_no_lib(self):
        """
        test gen_hash with no crypt library available
        """
        with self.assertRaises(SaltInvocationError):
            salt.utils.pycrypto.gen_hash()

    @patch("salt.utils.pycrypto.HAS_CRYPT", True)
    @patch("salt.utils.pycrypto.methods", {"crypt": None})
    @patch("salt.utils.pycrypto.HAS_PASSLIB", True)
    def test_gen_hash_selection(self):
        """
        verify the hash backend selection works correctly
        """
        with patch("salt.utils.pycrypto._gen_hash_crypt", autospec=True) as gh_crypt:
            with patch(
                "salt.utils.pycrypto._gen_hash_passlib", autospec=True
            ) as gh_passlib:
                with self.assertRaises(SaltInvocationError):
                    salt.utils.pycrypto.gen_hash(algorithm="doesntexist")

                salt.utils.pycrypto.gen_hash(algorithm="crypt")
                gh_crypt.assert_called_once()
                gh_passlib.assert_not_called()

                gh_crypt.reset_mock()
                salt.utils.pycrypto.gen_hash(algorithm="sha512")
                gh_crypt.assert_not_called()
                gh_passlib.assert_called_once()

    def test_gen_hash_crypt_warning(self):
        """
        Verify that a bad crypt salt triggers a warning
        """
        with patch("salt.utils.pycrypto.log", autospec=True) as log:
            try:
                salt.utils.pycrypto.gen_hash(
                    crypt_salt="toolong", password=self.passwd, algorithm="crypt"
                )
            except Exception:  # pylint: disable=broad-except
                pass
        log.warning.assert_called_with("Hash salt is too long for 'crypt' hash.")

    def test_secure_password(self):
        """
        test secure_password
        """
        ret = salt.utils.pycrypto.secure_password()
        check = re.compile(r"[!@#$%^&*()_=+]")
        self.assertIsNone(check.search(ret))
        self.assertTrue(ret)



# -*- coding: utf-8 -*-
"""
This is the a dummy proxy-minion designed for testing the proxy minion subsystem.
"""
from __future__ import absolute_import, print_function, unicode_literals

import logging

# Import python libs
import os
import pickle

# Import Salt libs
import salt.ext.six as six
import salt.utils.files

# This must be present or the Salt loader won't load this module
__proxyenabled__ = ["dummy"]


# Variables are scoped to this module so we can have persistent data
# across calls to fns in here.
DETAILS = {}

DETAILS["services"] = {"apache": "running", "ntp": "running", "samba": "stopped"}
DETAILS["packages"] = {
    "coreutils": "1.0",
    "apache": "2.4",
    "tinc": "1.4",
    "redbull": "999.99",
}
FILENAME = salt.utils.files.mkstemp()
# Want logging!
log = logging.getLogger(__file__)


# This does nothing, it's here just as an example and to provide a log
# entry when the module is loaded.
def __virtual__():
    """
    Only return if all the modules are available
    """
    log.debug("dummy proxy __virtual__() called...")
    return True


def _save_state(details):
    with salt.utils.files.fopen(FILENAME, "wb") as pck:
        pickle.dump(details, pck)


def _load_state():
    try:
        if six.PY3 is True:
            mode = "rb"
        else:
            mode = "r"

        with salt.utils.files.fopen(FILENAME, mode) as pck:
            DETAILS = pickle.load(pck)
    except EOFError:
        DETAILS = {}
        DETAILS["initialized"] = False
        _save_state(DETAILS)

    return DETAILS


# Every proxy module needs an 'init', though you can
# just put DETAILS['initialized'] = True here if nothing
# else needs to be done.


def init(opts):
    log.debug("dummy proxy init() called...")
    DETAILS["initialized"] = True
    _save_state(DETAILS)


def initialized():
    """
    Since grains are loaded in many different places and some of those
    places occur before the proxy can be initialized, return whether
    our init() function has been called
    """
    DETAILS = _load_state()
    return DETAILS.get("initialized", False)


def grains():
    """
    Make up some grains
    """
    DETAILS = _load_state()
    if "grains_cache" not in DETAILS:
        DETAILS["grains_cache"] = {
            "dummy_grain_1": "one",
            "dummy_grain_2": "two",
            "dummy_grain_3": "three",
        }
        _save_state(DETAILS)

    return DETAILS["grains_cache"]


def grains_refresh():
    """
    Refresh the grains
    """
    DETAILS = _load_state()
    DETAILS["grains_cache"] = None
    _save_state(DETAILS)
    return grains()


def fns():
    return {
        "details": "This key is here because a function in "
        "grains/rest_sample.py called fns() here in the proxymodule."
    }


def service_start(name):
    """
    Start a "service" on the dummy server
    """
    DETAILS = _load_state()
    DETAILS["services"][name] = "running"
    _save_state(DETAILS)
    return "running"


def service_stop(name):
    """
    Stop a "service" on the dummy server
    """
    DETAILS = _load_state()
    DETAILS["services"][name] = "stopped"
    _save_state(DETAILS)
    return "stopped"


def service_restart(name):
    """
    Restart a "service" on the REST server
    """
    return True


def service_list():
    """
    List "services" on the REST server
    """
    DETAILS = _load_state()
    return list(DETAILS["services"])


def service_status(name):
    """
    Check if a service is running on the REST server
    """
    DETAILS = _load_state()
    if DETAILS["services"][name] == "running":
        return {"comment": "running"}
    else:
        return {"comment": "stopped"}


def package_list():
    """
    List "packages" installed on the REST server
    """
    DETAILS = _load_state()
    return DETAILS["packages"]


def package_install(name, **kwargs):
    """
    Install a "package" on the REST server
    """
    DETAILS = _load_state()
    if kwargs.get("version", False):
        version = kwargs["version"]
    else:
        version = "1.0"
    DETAILS["packages"][name] = version
    _save_state(DETAILS)
    return {name: version}


def upgrade():
    """
    "Upgrade" packages
    """
    DETAILS = _load_state()
    pkgs = uptodate()
    DETAILS["packages"] = pkgs
    _save_state(DETAILS)
    return pkgs


def uptodate():
    """
    Call the REST endpoint to see if the packages on the "server" are up to date.
    """
    DETAILS = _load_state()
    for p in DETAILS["packages"]:
        version_float = float(DETAILS["packages"][p])
        version_float = version_float + 1.0
        DETAILS["packages"][p] = six.text_type(version_float)
    return DETAILS["packages"]


def package_remove(name):
    """
    Remove a "package" on the REST server
    """
    DETAILS = _load_state()
    DETAILS["packages"].pop(name)
    _save_state(DETAILS)
    return DETAILS["packages"]


def package_status(name):
    """
    Check the installation status of a package on the REST server
    """
    DETAILS = _load_state()
    if name in DETAILS["packages"]:
        return {name: DETAILS["packages"][name]}


def ping():
    """
    Degenerate ping
    """
    log.debug("dummy proxy returning ping")
    return True


def shutdown(opts):
    """
    For this proxy shutdown is a no-op
    """
    log.debug("dummy proxy shutdown() called...")
    DETAILS = _load_state()
    if "filename" in DETAILS:
        os.unlink(DETAILS["filename"])


def test_from_state():
    """
    Test function so we have something to call from a state
    :return:
    """
    log.debug("test_from_state called")
    return "testvalue"



# -*- coding: utf-8 -*-
"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.states.alternatives as alternatives

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class AlternativesTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.states.alternatives
    """

    def setup_loader_modules(self):
        return {alternatives: {}}

    # 'install' function tests: 1

    def test_install(self):
        """
        Test to install new alternative for defined <name>
        """
        name = "pager"
        link = "/usr/bin/pager"
        path = "/usr/bin/less"
        priority = 5

        ret = {
            "name": name,
            "link": link,
            "path": path,
            "priority": priority,
            "result": None,
            "changes": {},
            "comment": "",
        }

        bad_link = "/bin/pager"
        err = "the primary link for {0} must be {1}".format(name, link)

        mock_cinst = MagicMock(side_effect=[True, False])
        mock_cexist = MagicMock(
            side_effect=[True, False, False, True, False, False, False, True]
        )
        mock_out = MagicMock(side_effect=["", err, ""])
        mock_path = MagicMock(return_value=path)
        mock_link = MagicMock(return_value=link)
        with patch.dict(
            alternatives.__salt__,
            {
                "alternatives.check_installed": mock_cinst,
                "alternatives.check_exists": mock_cexist,
                "alternatives.install": mock_out,
                "alternatives.show_current": mock_path,
                "alternatives.show_link": mock_link,
            },
        ):
            comt = "Alternative {0} for {1} is already registered".format(path, name)
            ret.update({"comment": comt, "result": True})
            self.assertDictEqual(alternatives.install(name, link, path, priority), ret)

            comt = "Alternative will be set for {0} to {1} with priority {2}".format(
                name, path, priority
            )
            ret.update({"comment": comt, "result": None})
            with patch.dict(alternatives.__opts__, {"test": True}):
                self.assertDictEqual(
                    alternatives.install(name, link, path, priority), ret
                )

            comt = "Alternative for {0} set to path {1} with priority {2}".format(
                name, path, priority
            )
            ret.update(
                {
                    "comment": comt,
                    "result": True,
                    "changes": {
                        "name": name,
                        "link": link,
                        "path": path,
                        "priority": priority,
                    },
                }
            )
            with patch.dict(alternatives.__opts__, {"test": False}):
                self.assertDictEqual(
                    alternatives.install(name, link, path, priority), ret
                )

            comt = "Alternative for {0} not installed: {1}".format(name, err)
            ret.update(
                {"comment": comt, "result": False, "changes": {}, "link": bad_link}
            )
            with patch.dict(alternatives.__opts__, {"test": False}):
                self.assertDictEqual(
                    alternatives.install(name, bad_link, path, priority), ret
                )

            comt = "Alternative {0} for {1} registered with priority {2} and not set to default".format(
                path, name, priority
            )
            ret.update(
                {
                    "comment": comt,
                    "result": True,
                    "changes": {
                        "name": name,
                        "link": link,
                        "path": path,
                        "priority": priority,
                    },
                    "link": link,
                }
            )
            with patch.dict(alternatives.__opts__, {"test": False}):
                self.assertDictEqual(
                    alternatives.install(name, link, path, priority), ret
                )

    # 'remove' function tests: 1

    def test_remove(self):
        """
        Test to removes installed alternative for defined <name> and <path>
        or fallback to default alternative, if some defined before.
        """
        name = "pager"
        path = "/usr/bin/less"

        ret = {"name": name, "path": path, "result": None, "changes": {}, "comment": ""}

        mock = MagicMock(side_effect=[True, True, True, False, False])
        mock_bool = MagicMock(return_value=True)
        mock_show = MagicMock(side_effect=[False, True, True, False])
        with patch.dict(
            alternatives.__salt__,
            {
                "alternatives.check_exists": mock,
                "alternatives.show_current": mock_show,
                "alternatives.remove": mock_bool,
            },
        ):
            comt = "Alternative for {0} will be removed".format(name)
            ret.update({"comment": comt})
            with patch.dict(alternatives.__opts__, {"test": True}):
                self.assertDictEqual(alternatives.remove(name, path), ret)

            comt = "Alternative for {0} removed".format(name)
            ret.update({"comment": comt, "result": True})
            with patch.dict(alternatives.__opts__, {"test": False}):
                self.assertDictEqual(alternatives.remove(name, path), ret)

            comt = "Alternative for pager removed. Falling back to path True"
            ret.update({"comment": comt, "result": True, "changes": {"path": True}})
            with patch.dict(alternatives.__opts__, {"test": False}):
                self.assertDictEqual(alternatives.remove(name, path), ret)

            comt = "Alternative for {0} is set to it's default path True".format(name)
            ret.update({"comment": comt, "result": True, "changes": {}})
            self.assertDictEqual(alternatives.remove(name, path), ret)

            comt = "Alternative for {0} doesn't exist".format(name)
            ret.update({"comment": comt, "result": False})
            self.assertDictEqual(alternatives.remove(name, path), ret)

    # 'auto' function tests: 1

    def test_auto(self):
        """
        Test to instruct alternatives to use the highest priority
        path for <name>
        """
        name = "pager"

        ret = {"name": name, "result": True, "changes": {}, "comment": ""}

        mock = MagicMock(side_effect=[" auto mode", " ", " "])
        mock_auto = MagicMock(return_value=True)
        with patch.dict(
            alternatives.__salt__,
            {"alternatives.display": mock, "alternatives.auto": mock_auto},
        ):
            comt = "{0} already in auto mode".format(name)
            ret.update({"comment": comt})
            self.assertDictEqual(alternatives.auto(name), ret)

            comt = "{0} will be put in auto mode".format(name)
            ret.update({"comment": comt, "result": None})
            with patch.dict(alternatives.__opts__, {"test": True}):
                self.assertDictEqual(alternatives.auto(name), ret)

            ret.update({"comment": "", "result": True, "changes": {"result": True}})
            with patch.dict(alternatives.__opts__, {"test": False}):
                self.assertDictEqual(alternatives.auto(name), ret)

    # 'set_' function tests: 1

    def test_set(self):
        """
        Test to sets alternative for <name> to <path>, if <path> is defined
        as an alternative for <name>.
        """
        name = "pager"
        path = "/usr/bin/less"

        ret = {"name": name, "path": path, "result": True, "changes": {}, "comment": ""}

        mock = MagicMock(side_effect=[path, path, ""])
        mock_bool = MagicMock(return_value=True)
        mock_show = MagicMock(side_effect=[path, False, False, False, False])
        with patch.dict(
            alternatives.__salt__,
            {
                "alternatives.display": mock,
                "alternatives.show_current": mock_show,
                "alternatives.set": mock_bool,
            },
        ):
            comt = "Alternative for {0} already set to {1}".format(name, path)
            ret.update({"comment": comt})
            self.assertDictEqual(alternatives.set_(name, path), ret)

            comt = "Alternative for {0} will be set to path /usr/bin/less".format(name)
            ret.update({"comment": comt, "result": None})
            with patch.dict(alternatives.__opts__, {"test": True}):
                self.assertDictEqual(alternatives.set_(name, path), ret)

            comt = "Alternative for {0} not updated".format(name)
            ret.update({"comment": comt, "result": True})
            with patch.dict(alternatives.__opts__, {"test": False}):
                self.assertDictEqual(alternatives.set_(name, path), ret)

            comt = "Alternative {0} for {1} doesn't exist".format(path, name)
            ret.update({"comment": comt, "result": False})
            self.assertDictEqual(alternatives.set_(name, path), ret)



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import logging

import dateutil.parser as dateutil_parser
from tests.support.mock import MagicMock, patch
from tests.unit.utils.scheduler.base import SchedulerTestsBase

log = logging.getLogger(__name__)


class SchedulerMaxRunningTest(SchedulerTestsBase):
    """
    Validate the pkg module
    """

    def setUp(self):
        super(SchedulerMaxRunningTest, self).setUp()
        self.schedule.opts["loop_interval"] = 1

    def test_maxrunning_minion(self):
        """
        verify that scheduled job runs
        """
        self.schedule.opts["__role"] = "minion"

        job = {
            "schedule": {
                "maxrunning_minion": {
                    "function": "test.ping",
                    "seconds": 10,
                    "maxrunning": 1,
                }
            }
        }

        job_data = {
            "function": "test.ping",
            "run": True,
            "name": "maxrunning_minion",
            "seconds": 10,
            "_seconds": 10,
            "jid_include": True,
            "maxrunning": 1,
        }

        # Add the job to the scheduler
        self.schedule.opts.update(job)

        running_data = [
            {
                "fun_args": [],
                "jid": "20181018165923360935",
                "schedule": "maxrunning_minion",
                "pid": 15338,
                "fun": "test.ping",
                "id": "host",
            }
        ]

        run_time = dateutil_parser.parse("11/29/2017 4:00pm")

        with patch("salt.utils.minion.running", MagicMock(return_value=running_data)):
            with patch(
                "salt.utils.process.os_is_running", MagicMock(return_value=True)
            ):
                ret = self.schedule._check_max_running(
                    "test.ping", job_data, self.schedule.opts, now=run_time
                )
        self.assertIn("_skip_reason", ret)
        self.assertEqual("maxrunning", ret["_skip_reason"])
        self.assertEqual(False, ret["run"])

    def test_maxrunning_master(self):
        """
        verify that scheduled job runs
        """
        self.schedule.opts["__role"] = "master"

        job = {
            "schedule": {
                "maxrunning_master": {
                    "function": "state.orch",
                    "args": ["test.orch_test"],
                    "minutes": 1,
                    "maxrunning": 1,
                }
            }
        }

        job_data = {
            "function": "state.orch",
            "fun_args": ["test.orch_test"],
            "run": True,
            "name": "maxrunning_master",
            "minutes": 1,
            "jid_include": True,
            "maxrunning": 1,
        }

        # Add the job to the scheduler
        self.schedule.opts.update(job)

        running_data = [
            {
                "fun_args": ["test.orch_test"],
                "jid": "20181018165923360935",
                "schedule": "maxrunning_master",
                "pid": 15338,
                "fun": "state.orch",
                "id": "host",
            }
        ]

        run_time = dateutil_parser.parse("11/29/2017 4:00pm")

        with patch(
            "salt.utils.master.get_running_jobs", MagicMock(return_value=running_data)
        ):
            with patch(
                "salt.utils.process.os_is_running", MagicMock(return_value=True)
            ):
                ret = self.schedule._check_max_running(
                    "state.orch", job_data, self.schedule.opts, now=run_time
                )
        self.assertIn("_skip_reason", ret)
        self.assertEqual("maxrunning", ret["_skip_reason"])
        self.assertEqual(False, ret["run"])



# -*- coding: utf-8 -*-
"""
    :codeauthor: Alexander Pyatkin <asp@thexyz.net>
"""

# Import Python Libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.states.bower as bower
from salt.exceptions import CommandExecutionError

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class BowerTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.states.bower
    """

    def setup_loader_modules(self):
        return {bower: {"__opts__": {"test": False}}}

    def test_removed_not_installed(self):
        """
        Test if it returns True when specified package is not installed
        """

        mock = MagicMock(return_value={"underscore": {}})

        with patch.dict(bower.__salt__, {"bower.list": mock}):
            ret = bower.removed("jquery", "/path/to/project")
            expected = {
                "name": "jquery",
                "result": True,
                "comment": "Package 'jquery' is not installed",
                "changes": {},
            }
            self.assertEqual(ret, expected)

    def test_removed_with_error(self):
        """
        Test if returns False when list packages fails
        """

        mock = MagicMock(side_effect=CommandExecutionError)

        with patch.dict(bower.__salt__, {"bower.list": mock}):
            ret = bower.removed("underscore", "/path/to/project")
            expected = {
                "name": "underscore",
                "result": False,
                "comment": "Error removing 'underscore': ",
                "changes": {},
            }
            self.assertEqual(ret, expected)

    def test_removed_existing(self):
        """
        Test if it returns True when specified package is installed and
        uninstall succeeds
        """

        mock_list = MagicMock(return_value={"underscore": {}})
        mock_uninstall = MagicMock(return_value=True)

        with patch.dict(
            bower.__salt__, {"bower.list": mock_list, "bower.uninstall": mock_uninstall}
        ):
            ret = bower.removed("underscore", "/path/to/project")
            expected = {
                "name": "underscore",
                "result": True,
                "comment": "Package 'underscore' was successfully removed",
                "changes": {"underscore": "Removed"},
            }
            self.assertEqual(ret, expected)

    def test_removed_existing_with_error(self):
        """
        Test if it returns False when specified package is installed and
        uninstall fails
        """

        mock_list = MagicMock(return_value={"underscore": {}})
        mock_uninstall = MagicMock(side_effect=CommandExecutionError)

        with patch.dict(
            bower.__salt__, {"bower.list": mock_list, "bower.uninstall": mock_uninstall}
        ):
            ret = bower.removed("underscore", "/path/to/project")
            expected = {
                "name": "underscore",
                "result": False,
                "comment": "Error removing 'underscore': ",
                "changes": {},
            }
            self.assertEqual(ret, expected)

    def test_bootstrap_with_error(self):
        """
        Test if it return False when install packages fails
        """

        mock = MagicMock(side_effect=CommandExecutionError)

        with patch.dict(bower.__salt__, {"bower.install": mock}):
            ret = bower.bootstrap("/path/to/project")
            expected = {
                "name": "/path/to/project",
                "result": False,
                "comment": "Error bootstrapping '/path/to/project': ",
                "changes": {},
            }
            self.assertEqual(ret, expected)

    def test_bootstrap_not_needed(self):
        """
        Test if it returns True when there is nothing to install
        """

        mock = MagicMock(return_value=False)

        with patch.dict(bower.__salt__, {"bower.install": mock}):
            ret = bower.bootstrap("/path/to/project")
            expected = {
                "name": "/path/to/project",
                "result": True,
                "comment": "Directory is already bootstrapped",
                "changes": {},
            }
            self.assertEqual(ret, expected)

    def test_bootstrap_success(self):
        """
        Test if it returns True when install packages succeeds
        """

        mock = MagicMock(return_value=True)

        with patch.dict(bower.__salt__, {"bower.install": mock}):
            ret = bower.bootstrap("/path/to/project")
            expected = {
                "name": "/path/to/project",
                "result": True,
                "comment": "Directory was successfully bootstrapped",
                "changes": {"/path/to/project": "Bootstrapped"},
            }
            self.assertEqual(ret, expected)

    def test_installed_with_error(self):
        """
        Test if it returns False when list packages fails
        """

        mock = MagicMock(side_effect=CommandExecutionError)

        with patch.dict(bower.__salt__, {"bower.list": mock}):
            ret = bower.installed("underscore", "/path/to/project")
            expected = {
                "name": "underscore",
                "result": False,
                "comment": "Error looking up 'underscore': ",
                "changes": {},
            }
            self.assertEqual(ret, expected)

    def test_installed_not_needed(self):
        """
        Test if it returns True when there is nothing to install
        """

        mock = MagicMock(
            return_value={
                "underscore": {"pkgMeta": {"version": "1.7.0"}},
                "jquery": {"pkgMeta": {"version": "2.0.0"}},
            }
        )

        with patch.dict(bower.__salt__, {"bower.list": mock}):
            ret = bower.installed(
                "test", "/path/to/project", ["underscore", "jquery#2.0.0"]
            )
            expected = {
                "name": "test",
                "result": True,
                "comment": (
                    "Package(s) 'underscore, jquery#2.0.0'"
                    " satisfied by underscore#1.7.0, jquery#2.0.0"
                ),
                "changes": {},
            }
            self.assertEqual(ret, expected)

    def test_installed_new_with_exc(self):
        """
        Test if it returns False when install packages fails (exception)
        """

        mock_list = MagicMock(return_value={})
        mock_install = MagicMock(side_effect=CommandExecutionError)

        with patch.dict(
            bower.__salt__, {"bower.list": mock_list, "bower.install": mock_install}
        ):
            ret = bower.installed("underscore", "/path/to/project")
            expected = {
                "name": "underscore",
                "result": False,
                "comment": "Error installing 'underscore': ",
                "changes": {},
            }
            self.assertEqual(ret, expected)

    def test_installed_new_with_error(self):
        """
        Test if returns False when install packages fails (bower error)
        """

        mock_list = MagicMock(return_value={})
        mock_install = MagicMock(return_value=False)

        with patch.dict(
            bower.__salt__, {"bower.list": mock_list, "bower.install": mock_install}
        ):
            ret = bower.installed("underscore", "/path/to/project")
            expected = {
                "name": "underscore",
                "result": False,
                "comment": "Could not install package(s) 'underscore'",
                "changes": {},
            }
            self.assertEqual(ret, expected)

    def test_installed_success(self):
        """
        Test if it returns True when install succeeds
        """

        mock_list = MagicMock(return_value={})
        mock_install = MagicMock(return_value=True)

        with patch.dict(
            bower.__salt__, {"bower.list": mock_list, "bower.install": mock_install}
        ):
            ret = bower.installed("underscore", "/path/to/project")
            expected = {
                "name": "underscore",
                "result": True,
                "comment": "Package(s) 'underscore' successfully installed",
                "changes": {"new": ["underscore"], "old": []},
            }
            self.assertEqual(ret, expected)



# -*- coding: utf-8 -*-
# Copyright (c) 2018 Cisco and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Util functions for the NXOS modules.
"""
from __future__ import absolute_import, print_function, unicode_literals

import collections
import http.client
import json
import logging
import os
import re
import socket
from collections.abc import Iterable

import salt.utils.http
from salt.exceptions import (
    CommandExecutionError,
    NxosClientError,
    NxosError,
    NxosRequestNotSupported,
)
from salt.utils.args import clean_kwargs

log = logging.getLogger(__name__)


class UHTTPConnection(http.client.HTTPConnection):
    """
    Subclass of Python library HTTPConnection that uses a unix-domain socket.
    """

    def __init__(self, path):
        http.client.HTTPConnection.__init__(self, "localhost")
        self.path = path

    def connect(self):
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.connect(self.path)
        self.sock = sock


class NxapiClient(object):
    """
    Class representing an NX-API client that connects over http(s) or
    unix domain socket (UDS).
    """

    # Location of unix domain socket for NX-API localhost
    NXAPI_UDS = "/tmp/nginx_local/nginx_1_be_nxapi.sock"
    # NXAPI listens for remote connections to "http(s)://<switch IP>/ins"
    # NXAPI listens for local connections to "http(s)://<UDS>/ins_local"
    NXAPI_REMOTE_URI_PATH = "/ins"
    NXAPI_UDS_URI_PATH = "/ins_local"
    NXAPI_VERSION = "1.0"

    def __init__(self, **nxos_kwargs):
        """
        Initialize NxapiClient() connection object.  By default this connects
        to the local unix domain socket (UDS).  If http(s) is required to
        connect to a remote device then
            nxos_kwargs['host'],
            nxos_kwargs['username'],
            nxos_kwargs['password'],
            nxos_kwargs['transport'],
            nxos_kwargs['port'],
        parameters must be provided.
        """
        self.nxargs = self._prepare_conn_args(clean_kwargs(**nxos_kwargs))
        # Default: Connect to unix domain socket on localhost.
        if self.nxargs["connect_over_uds"]:
            if not os.path.exists(self.NXAPI_UDS):
                raise NxosClientError(
                    "No host specified and no UDS found at {0}\n".format(self.NXAPI_UDS)
                )

            # Create UHTTPConnection object for NX-API communication over UDS.
            log.info("Nxapi connection arguments: %s", self.nxargs)
            log.info("Connecting over unix domain socket")
            self.connection = UHTTPConnection(self.NXAPI_UDS)
        else:
            # Remote connection - Proxy Minion, connect over http(s)
            log.info("Nxapi connection arguments: %s", self.nxargs)
            log.info("Connecting over %s", self.nxargs["transport"])
            self.connection = salt.utils.http.query

    def _use_remote_connection(self, kwargs):
        """
        Determine if connection is local or remote
        """
        kwargs["host"] = kwargs.get("host")
        kwargs["username"] = kwargs.get("username")
        kwargs["password"] = kwargs.get("password")
        if (
            kwargs["host"] is None
            or kwargs["username"] is None
            or kwargs["password"] is None
        ):
            return False
        else:
            return True

    def _prepare_conn_args(self, kwargs):
        """
        Set connection arguments for remote or local connection.
        """
        kwargs["connect_over_uds"] = True
        kwargs["timeout"] = kwargs.get("timeout", 60)
        kwargs["cookie"] = kwargs.get("cookie", "admin")
        if self._use_remote_connection(kwargs):
            kwargs["transport"] = kwargs.get("transport", "https")
            if kwargs["transport"] == "https":
                kwargs["port"] = kwargs.get("port", 443)
            else:
                kwargs["port"] = kwargs.get("port", 80)
            kwargs["verify"] = kwargs.get("verify", True)
            if isinstance(kwargs["verify"], bool):
                kwargs["verify_ssl"] = kwargs["verify"]
            else:
                kwargs["ca_bundle"] = kwargs["verify"]
            kwargs["connect_over_uds"] = False
        return kwargs

    def _build_request(self, type, commands):
        """
        Build NX-API JSON request.
        """
        request = {}
        headers = {
            "content-type": "application/json",
        }
        if self.nxargs["connect_over_uds"]:
            user = self.nxargs["cookie"]
            headers["cookie"] = "nxapi_auth=" + user + ":local"
            request["url"] = self.NXAPI_UDS_URI_PATH
        else:
            request["url"] = "{transport}://{host}:{port}{uri}".format(
                transport=self.nxargs["transport"],
                host=self.nxargs["host"],
                port=self.nxargs["port"],
                uri=self.NXAPI_REMOTE_URI_PATH,
            )

        if isinstance(commands, (list, set, tuple)):
            commands = " ; ".join(commands)
        payload = {}
        # Some versions of NX-OS fail to process the payload properly if
        # 'input' gets serialized before 'type' and the payload of 'input'
        # contains the string 'type'.  Use an ordered dict to enforce ordering.
        payload["ins_api"] = collections.OrderedDict()
        payload["ins_api"]["version"] = self.NXAPI_VERSION
        payload["ins_api"]["type"] = type
        payload["ins_api"]["chunk"] = "0"
        payload["ins_api"]["sid"] = "1"
        payload["ins_api"]["input"] = commands
        payload["ins_api"]["output_format"] = "json"

        request["headers"] = headers
        request["payload"] = json.dumps(payload)
        request["opts"] = {"http_request_timeout": self.nxargs["timeout"]}
        log.info("request: %s", request)
        return request

    def request(self, type, command_list):
        """
        Send NX-API JSON request to the NX-OS device.
        """
        req = self._build_request(type, command_list)
        if self.nxargs["connect_over_uds"]:
            self.connection.request("POST", req["url"], req["payload"], req["headers"])
            response = self.connection.getresponse()
        else:
            response = self.connection(
                req["url"],
                method="POST",
                opts=req["opts"],
                data=req["payload"],
                header_dict=req["headers"],
                decode=True,
                decode_type="json",
                **self.nxargs
            )

        return self.parse_response(response, command_list)

    def parse_response(self, response, command_list):
        """
        Parse NX-API JSON response from the NX-OS device.
        """
        # Check for 500 level NX-API Server Errors
        if isinstance(response, Iterable) and "status" in response:
            if int(response["status"]) >= 500:
                raise NxosError("{}".format(response))
            else:
                raise NxosError("NX-API Request Not Supported: {}".format(response))

        if isinstance(response, Iterable):
            body = response["dict"]
        else:
            body = response

        if self.nxargs["connect_over_uds"]:
            body = json.loads(response.read().decode("utf-8"))

        # Proceed with caution.  The JSON may not be complete.
        # Don't just return body['ins_api']['outputs']['output'] directly.
        output = body.get("ins_api")
        if output is None:
            raise NxosClientError("Unexpected JSON output\n{0}".format(body))
        if output.get("outputs"):
            output = output["outputs"]
        if output.get("output"):
            output = output["output"]

        # The result list stores results for each command that was sent to
        # nxapi.
        result = []
        # Keep track of successful commands using previous_commands list so
        # they can be displayed if a specific command fails in a chain of
        # commands.
        previous_commands = []

        # Make sure output and command_list lists to be processed in the
        # subesequent loop.
        if not isinstance(output, list):
            output = [output]
        if not isinstance(command_list, list):
            command_list = [command_list]
        if len(command_list) == 1 and ";" in command_list[0]:
            command_list = [cmd.strip() for cmd in command_list[0].split(";")]

        for cmd_result, cmd in zip(output, command_list):
            code = cmd_result.get("code")
            msg = cmd_result.get("msg")
            log.info("command %s:", cmd)
            log.info("PARSE_RESPONSE: %s %s", code, msg)
            if code == "400":
                raise CommandExecutionError(
                    {
                        "rejected_input": cmd,
                        "code": code,
                        "message": msg,
                        "cli_error": cmd_result.get("clierror"),
                        "previous_commands": previous_commands,
                    }
                )
            elif code == "413":
                raise NxosRequestNotSupported("Error 413: {}".format(msg))
            elif code != "200":
                raise NxosError("Unknown Error: {}, Code: {}".format(msg, code))
            else:
                previous_commands.append(cmd)
                result.append(cmd_result["body"])

        return result


def nxapi_request(commands, method="cli_show", **kwargs):
    """
    Send exec and config commands to the NX-OS device over NX-API.

    commands
        The exec or config commands to be sent.

    method:
        ``cli_show_ascii``: Return raw test or unstructured output.
        ``cli_show``: Return structured output.
        ``cli_conf``: Send configuration commands to the device.
        Defaults to ``cli_show``.

    transport: ``https``
        Specifies the type of connection transport to use. Valid values for the
        connection are ``http``, and  ``https``.

    host: ``localhost``
        The IP address or DNS host name of the device.

    username: ``admin``
        The username to pass to the device to authenticate the NX-API connection.

    password
        The password to pass to the device to authenticate the NX-API connection.

    port
        The TCP port of the endpoint for the NX-API connection. If this keyword is
        not specified, the default value is automatically determined by the
        transport type (``80`` for ``http``, or ``443`` for ``https``).

    timeout: ``60``
        Time in seconds to wait for the device to respond. Default: 60 seconds.

    verify: ``True``
        Either a boolean, in which case it controls whether we verify the NX-API
        TLS certificate, or a string, in which case it must be a path to a CA bundle
        to use. Defaults to ``True``.
    """
    client = NxapiClient(**kwargs)
    return client.request(method, commands)


def ping(**kwargs):
    """
    Verify connection to the NX-OS device over UDS.
    """
    return NxapiClient(**kwargs).nxargs["connect_over_uds"]


# Grains Functions


def _parser(block):
    return re.compile("^{block}\n(?:^[ \n].*$\n?)+".format(block=block), re.MULTILINE)


def _parse_software(data):
    """
    Internal helper function to parse sotware grain information.
    """
    ret = {"software": {}}
    software = _parser("Software").search(data).group(0)
    matcher = re.compile("^  ([^:]+): *([^\n]+)", re.MULTILINE)
    for line in matcher.finditer(software):
        key, val = line.groups()
        ret["software"][key] = val
    return ret["software"]


def _parse_hardware(data):
    """
    Internal helper function to parse hardware grain information.
    """
    ret = {"hardware": {}}
    hardware = _parser("Hardware").search(data).group(0)
    matcher = re.compile("^  ([^:\n]+): *([^\n]+)", re.MULTILINE)
    for line in matcher.finditer(hardware):
        key, val = line.groups()
        ret["hardware"][key] = val
    return ret["hardware"]


def _parse_plugins(data):
    """
    Internal helper function to parse plugin grain information.
    """
    ret = {"plugins": []}
    plugins = _parser("plugin").search(data).group(0)
    matcher = re.compile("^  (?:([^,]+), )+([^\n]+)", re.MULTILINE)
    for line in matcher.finditer(plugins):
        ret["plugins"].extend(line.groups())
    return ret["plugins"]


def version_info():
    client = NxapiClient()
    return client.request("cli_show_ascii", "show version")[0]


def system_info(data):
    """
    Helper method to return parsed system_info
    from the 'show version' command.
    """
    if not data:
        return {}
    info = {
        "software": _parse_software(data),
        "hardware": _parse_hardware(data),
        "plugins": _parse_plugins(data),
    }
    return {"nxos": info}



# -*- coding: utf-8 -*-
"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.states.chef as chef

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class ChefTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.states.chef
    """

    def setup_loader_modules(self):
        return {chef: {}}

    # 'client' function tests: 1

    def test_client(self):
        """
        Test to run chef-client
        """
        name = "my-chef-run"

        ret = {"name": name, "result": False, "changes": {}, "comment": ""}

        mock = MagicMock(return_value={"retcode": 1, "stdout": "", "stderr": "error"})
        with patch.dict(chef.__salt__, {"chef.client": mock}):
            with patch.dict(chef.__opts__, {"test": True}):
                comt = "\nerror"
                ret.update({"comment": comt})
                self.assertDictEqual(chef.client(name), ret)

    # 'solo' function tests: 1

    def test_solo(self):
        """
        Test to run chef-solo
        """
        name = "my-chef-run"

        ret = {"name": name, "result": False, "changes": {}, "comment": ""}

        mock = MagicMock(return_value={"retcode": 1, "stdout": "", "stderr": "error"})
        with patch.dict(chef.__salt__, {"chef.solo": mock}):
            with patch.dict(chef.__opts__, {"test": True}):
                comt = "\nerror"
                ret.update({"comment": comt})
                self.assertDictEqual(chef.solo(name), ret)



# -*- coding: utf-8 -*-

# Import python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.states.smartos as smartos
from salt.utils.odict import OrderedDict

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import patch
from tests.support.unit import TestCase


class SmartOsTestCase(TestCase, LoaderModuleMockMixin):
    """
    TestCase for salt.states.smartos
    """

    def setup_loader_modules(self):
        return {smartos: {"__opts__": {"test": False}}}

    def test_config_present_does_not_exist(self):
        """
        Test salt.states.smartos.config_present
        when the config files does not exist
        """
        name = "test"
        value = "test_value"
        with patch("os.path.isfile", return_value=False):
            with patch("salt.utils.atomicfile.atomic_open", side_effect=IOError):
                ret = smartos.config_present(name=name, value=value)
        assert not ret["result"]
        assert ret[
            "comment"
        ] == 'Could not add property {0} with value "{1}" to config'.format(name, value)

    def test_parse_vmconfig_vrrp(self):
        """
        Test _parse_vmconfig's vrid -> mac convertor

        SmartOS will always use a mac based on the vrrp_vrid,
        so we will replace the provided mac with the one based
        on this value.

        Doing so ensures that 'old' nics are removed and 'new'
        nics get added as these actions are keyed on the mac
        property.
        """
        # NOTE: vmconfig is not a full vmadm payload,
        #       this is not an issue given we are only testing
        #       the vrrp_vrid to mac conversions
        ret = smartos._parse_vmconfig(
            OrderedDict(
                [
                    (
                        "nics",
                        OrderedDict(
                            [
                                (
                                    "00:00:5e:00:01:01",
                                    OrderedDict(
                                        [
                                            ("vrrp_vrid", 1),
                                            ("vrrp_primary_ip", "12.34.5.6"),
                                        ]
                                    ),
                                ),
                                (
                                    "00:00:5e:00:01:24",
                                    OrderedDict(
                                        [
                                            ("vrrp_vrid", 240),
                                            ("vrrp_primary_ip", "12.34.5.6"),
                                        ]
                                    ),
                                ),
                                (
                                    "00:22:06:00:00:01",
                                    OrderedDict([("ips", ["12.34.5.6/24"])]),
                                ),
                            ]
                        ),
                    )
                ]
            ),
            {"nics": "mac"},
        )

        # NOTE: nics.0 is a vrrp nic with correct mac (check mac == vrid based -> unchanged)
        assert ret["nics"][0]["mac"] == "00:00:5e:00:01:01"

        # NOTE: nics.1 is a vrrp nic with incorrect mac (check mac == vrid based -> changed)
        assert ret["nics"][1]["mac"] == "00:00:5e:00:01:f0"

        # NOTE: nics.2 was not a vrrp nic (check mac was not changed)
        assert ret["nics"][2]["mac"] == "00:22:06:00:00:01"



# -*- coding: utf-8 -*-
"""
    :codeauthor: Mike Place <mp@saltstack.com>
"""

# Import python libs
from __future__ import absolute_import, print_function, unicode_literals

import errno
import logging
import os
import socket
import threading

import salt.config
import salt.exceptions
import salt.ext.tornado.gen
import salt.ext.tornado.ioloop
import salt.ext.tornado.testing
import salt.transport.client
import salt.transport.ipc
import salt.transport.server
import salt.utils.platform
from salt.ext import six
from salt.ext.six.moves import range
from tests.support.mock import MagicMock

# Import Salt Testing libs
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import skipIf

log = logging.getLogger(__name__)


@skipIf(salt.utils.platform.is_windows(), "Windows does not support Posix IPC")
class BaseIPCReqCase(salt.ext.tornado.testing.AsyncTestCase):
    """
    Test the req server/client pair
    """

    def setUp(self):
        super(BaseIPCReqCase, self).setUp()
        # self._start_handlers = dict(self.io_loop._handlers)
        self.socket_path = os.path.join(RUNTIME_VARS.TMP, "ipc_test.ipc")

        self.server_channel = salt.transport.ipc.IPCMessageServer(
            self.socket_path,
            io_loop=self.io_loop,
            payload_handler=self._handle_payload,
        )
        self.server_channel.start()

        self.payloads = []

    def tearDown(self):
        super(BaseIPCReqCase, self).tearDown()
        # failures = []
        try:
            self.server_channel.close()
        except socket.error as exc:
            if exc.errno != errno.EBADF:
                # If its not a bad file descriptor error, raise
                raise
        os.unlink(self.socket_path)
        # for k, v in six.iteritems(self.io_loop._handlers):
        #    if self._start_handlers.get(k) != v:
        #        failures.append((k, v))
        # if len(failures) > 0:
        #    raise Exception('FDs still attached to the IOLoop: {0}'.format(failures))
        del self.payloads
        del self.socket_path
        del self.server_channel
        # del self._start_handlers

    @salt.ext.tornado.gen.coroutine
    def _handle_payload(self, payload, reply_func):
        self.payloads.append(payload)
        yield reply_func(payload)
        if isinstance(payload, dict) and payload.get("stop"):
            self.stop()


class IPCMessageClient(BaseIPCReqCase):
    """
    Test all of the clear msg stuff
    """

    def _get_channel(self):
        if not hasattr(self, "channel") or self.channel is None:
            self.channel = salt.transport.ipc.IPCMessageClient(
                socket_path=self.socket_path, io_loop=self.io_loop,
            )
            self.channel.connect(callback=self.stop)
            self.wait()
        return self.channel

    def setUp(self):
        super(IPCMessageClient, self).setUp()
        self.channel = self._get_channel()

    def tearDown(self):
        super(IPCMessageClient, self).tearDown()
        try:
            # Make sure we close no matter what we've done in the tests
            del self.channel
        except socket.error as exc:
            if exc.errno != errno.EBADF:
                # If its not a bad file descriptor error, raise
                raise
        finally:
            self.channel = None

    def test_singleton(self):
        channel = self._get_channel()
        assert self.channel is channel
        # Delete the local channel. Since there's still one more refefence
        # __del__ wasn't called
        del channel
        assert self.channel
        msg = {"foo": "bar", "stop": True}
        self.channel.send(msg)
        self.wait()
        self.assertEqual(self.payloads[0], msg)

    def test_basic_send(self):
        msg = {"foo": "bar", "stop": True}
        self.channel.send(msg)
        self.wait()
        self.assertEqual(self.payloads[0], msg)

    def test_many_send(self):
        msgs = []
        self.server_channel.stream_handler = MagicMock()

        for i in range(0, 1000):
            msgs.append("test_many_send_{0}".format(i))

        for i in msgs:
            self.channel.send(i)
        self.channel.send({"stop": True})
        self.wait()
        self.assertEqual(self.payloads[:-1], msgs)

    def test_very_big_message(self):
        long_str = "".join([six.text_type(num) for num in range(10 ** 5)])
        msg = {"long_str": long_str, "stop": True}
        self.channel.send(msg)
        self.wait()
        self.assertEqual(msg, self.payloads[0])

    def test_multistream_sends(self):
        local_channel = self._get_channel()

        for c in (self.channel, local_channel):
            c.send("foo")

        self.channel.send({"stop": True})
        self.wait()
        self.assertEqual(self.payloads[:-1], ["foo", "foo"])

    def test_multistream_errors(self):
        local_channel = self._get_channel()

        for c in (self.channel, local_channel):
            c.send(None)

        for c in (self.channel, local_channel):
            c.send("foo")

        self.channel.send({"stop": True})
        self.wait()
        self.assertEqual(self.payloads[:-1], [None, None, "foo", "foo"])


@skipIf(salt.utils.platform.is_windows(), "Windows does not support Posix IPC")
class IPCMessagePubSubCase(salt.ext.tornado.testing.AsyncTestCase):
    """
    Test all of the clear msg stuff
    """

    def setUp(self):
        super(IPCMessagePubSubCase, self).setUp()
        self.opts = {"ipc_write_buffer": 0}
        self.socket_path = os.path.join(RUNTIME_VARS.TMP, "ipc_test.ipc")
        self.pub_channel = self._get_pub_channel()
        self.sub_channel = self._get_sub_channel()

    def _get_pub_channel(self):
        pub_channel = salt.transport.ipc.IPCMessagePublisher(
            self.opts, self.socket_path,
        )
        pub_channel.start()
        return pub_channel

    def _get_sub_channel(self):
        sub_channel = salt.transport.ipc.IPCMessageSubscriber(
            socket_path=self.socket_path, io_loop=self.io_loop,
        )
        sub_channel.connect(callback=self.stop)
        self.wait()
        return sub_channel

    def tearDown(self):
        super(IPCMessagePubSubCase, self).tearDown()
        try:
            self.pub_channel.close()
        except socket.error as exc:
            if exc.errno != errno.EBADF:
                # If its not a bad file descriptor error, raise
                raise
        try:
            self.sub_channel.close()
        except socket.error as exc:
            if exc.errno != errno.EBADF:
                # If its not a bad file descriptor error, raise
                raise
        os.unlink(self.socket_path)
        del self.pub_channel
        del self.sub_channel

    def test_multi_client_reading(self):
        # To be completely fair let's create 2 clients.
        client1 = self.sub_channel
        client2 = self._get_sub_channel()
        call_cnt = []

        # Create a watchdog to be safe from hanging in sync loops (what old code did)
        evt = threading.Event()

        def close_server():
            if evt.wait(1):
                return
            client2.close()
            self.stop()

        watchdog = threading.Thread(target=close_server)
        watchdog.start()

        # Runs in ioloop thread so we're safe from race conditions here
        def handler(raw):
            call_cnt.append(raw)
            if len(call_cnt) >= 2:
                evt.set()
                self.stop()

        # Now let both waiting data at once
        client1.read_async(handler)
        client2.read_async(handler)
        self.pub_channel.publish("TEST")
        self.wait()
        self.assertEqual(len(call_cnt), 2)
        self.assertEqual(call_cnt[0], "TEST")
        self.assertEqual(call_cnt[1], "TEST")

    def test_sync_reading(self):
        # To be completely fair let's create 2 clients.
        client1 = self.sub_channel
        client2 = self._get_sub_channel()
        call_cnt = []

        # Now let both waiting data at once
        self.pub_channel.publish("TEST")
        ret1 = client1.read_sync()
        ret2 = client2.read_sync()
        self.assertEqual(ret1, "TEST")
        self.assertEqual(ret2, "TEST")



# -*- coding: utf-8 -*-
"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.states.postgres_database as postgres_database

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class PostgresDatabaseTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.states.postgres_database
    """

    def setup_loader_modules(self):
        return {postgres_database: {}}

    # 'present' function tests: 1

    def test_present(self):
        """
        Test to ensure that the named database is present
        with the specified properties.
        """
        name = "frank"

        ret = {"name": name, "changes": {}, "result": False, "comment": ""}

        mock_t = MagicMock(return_value=True)
        mock = MagicMock(return_value={name: {}})
        with patch.dict(
            postgres_database.__salt__,
            {"postgres.db_list": mock, "postgres.db_alter": mock_t},
        ):
            comt = "Database {0} is already present".format(name)
            ret.update({"comment": comt, "result": True})
            self.assertDictEqual(postgres_database.present(name), ret)

            comt = (
                "Database frank has wrong parameters "
                "which couldn't be changed on fly."
            )
            ret.update({"comment": comt, "result": False})
            self.assertDictEqual(
                postgres_database.present(name, tablespace="A", lc_collate=True), ret
            )

            with patch.dict(postgres_database.__opts__, {"test": True}):
                comt = "Database frank exists, " "but parameters need to be changed"
                ret.update({"comment": comt, "result": None})
                self.assertDictEqual(
                    postgres_database.present(name, tablespace="A"), ret
                )

            with patch.dict(postgres_database.__opts__, {"test": False}):
                comt = "Parameters for database frank have been changed"
                ret.update(
                    {
                        "comment": comt,
                        "result": True,
                        "changes": {name: "Parameters changed"},
                    }
                )
                self.assertDictEqual(
                    postgres_database.present(name, tablespace="A"), ret
                )

    # 'absent' function tests: 1

    def test_absent(self):
        """
        Test to ensure that the named database is absent.
        """
        name = "frank"

        ret = {"name": name, "changes": {}, "result": False, "comment": ""}

        mock_t = MagicMock(return_value=True)
        mock = MagicMock(side_effect=[True, True, False])
        with patch.dict(
            postgres_database.__salt__,
            {"postgres.db_exists": mock, "postgres.db_remove": mock_t},
        ):
            with patch.dict(postgres_database.__opts__, {"test": True}):
                comt = "Database {0} is set to be removed".format(name)
                ret.update({"comment": comt, "result": None})
                self.assertDictEqual(postgres_database.absent(name), ret)

            with patch.dict(postgres_database.__opts__, {"test": False}):
                comt = "Database {0} has been removed".format(name)
                ret.update(
                    {"comment": comt, "result": True, "changes": {name: "Absent"}}
                )
                self.assertDictEqual(postgres_database.absent(name), ret)

                comt = "Database {0} is not present, so it cannot be removed".format(
                    name
                )
                ret.update({"comment": comt, "result": True, "changes": {}})
                self.assertDictEqual(postgres_database.absent(name), ret)



# -*- coding: utf-8 -*-
"""
:codeauthor: Herbert Buurman <herbert.buurman@ogd.nl>
"""

# Import future libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Module
import salt.states.acme as acme

# Import Salt Testing libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class AcmeTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.acme
    """

    def setup_loader_modules(self):
        return {acme: {"__opts__": {"test": False}}}

    def test_cert_no_changes_t(self):
        """
        Test cert state with no needed changes. (test=True)
        """
        # With test=True
        with patch.dict(
            acme.__salt__,
            {  # pylint: disable=no-member
                "acme.has": MagicMock(return_value=True),
                "acme.needs_renewal": MagicMock(return_value=False),
            },
        ), patch.dict(
            acme.__opts__, {"test": True}
        ):  # pylint: disable=no-member
            self.assertEqual(
                acme.cert("test"),
                {
                    "name": "test",
                    "result": True,
                    "comment": ["Certificate test exists and does not need renewal."],
                    "changes": {},
                },
            )

    def test_cert_no_changes(self):
        """
        Test cert state with no needed changes. (test=False)
        """
        # With test=False
        with patch.dict(
            acme.__salt__,
            {  # pylint: disable=no-member
                "acme.has": MagicMock(return_value=True),
                "acme.needs_renewal": MagicMock(return_value=False),
            },
        ):
            self.assertEqual(
                acme.cert("test"),
                {
                    "name": "test",
                    "result": True,
                    "comment": ["Certificate test exists and does not need renewal."],
                    "changes": {},
                },
            )

    def test_cert_fresh_certificate_t(self):
        """
        Test cert state fetching a new certificate. (test=True)
        """
        # With test=True
        with patch.dict(
            acme.__salt__,
            {  # pylint: disable=no-member
                "acme.has": MagicMock(return_value=False),
                # 'acme.cert': MagicMock(return_value={'result': True, 'comment': 'Mockery'}),
                "acme.info": MagicMock(return_value={"foo": "bar"}),
            },
        ), patch.dict(
            acme.__opts__, {"test": True}
        ):  # pylint: disable=no-member
            self.assertEqual(
                acme.cert("test"),
                {
                    "name": "test",
                    "result": None,
                    "comment": ["Certificate test would have been obtained."],
                    "changes": {"old": "current certificate", "new": "new certificate"},
                },
            )

    def test_cert_fresh_certificate(self):
        """
        Test cert state fetching a new certificate. (test=False)
        """
        # With test=False
        with patch.dict(
            acme.__salt__,
            {  # pylint: disable=no-member
                "acme.has": MagicMock(return_value=False),
                "acme.cert": MagicMock(
                    return_value={"result": True, "comment": "Mockery"}
                ),
                "acme.info": MagicMock(return_value={"foo": "bar"}),
            },
        ):
            self.assertEqual(
                acme.cert("test"),
                {
                    "name": "test",
                    "result": True,
                    "comment": ["Mockery"],
                    "changes": {"new": {"foo": "bar"}},
                },
            )

    def test_cert_renew_certificate_t(self):
        """
        Test cert state renewing a certificate. (test=True)
        """
        with patch.dict(
            acme.__salt__,
            {  # pylint: disable=no-member
                "acme.has": MagicMock(return_value=True),
                "acme.needs_renewal": MagicMock(return_value=True),
                "acme.info": MagicMock(
                    side_effect=[{"name": "old cert"}, {"name": "new cert"}]
                ),
                "acme.cert": MagicMock(
                    return_value={"result": True, "comment": "Mockery"}
                ),
            },
        ), patch.dict(
            acme.__opts__, {"test": True}
        ):  # pylint: disable=no-member
            self.assertEqual(
                acme.cert("test"),
                {
                    "name": "test",
                    "result": None,
                    "comment": ["Certificate test would have been renewed."],
                    "changes": {"old": "current certificate", "new": "new certificate"},
                },
            )

    def test_cert_renew_certificate(self):
        """
        Test cert state renewing a certificate. (test=False)
        """
        with patch.dict(
            acme.__salt__,
            {  # pylint: disable=no-member
                "acme.has": MagicMock(return_value=True),
                "acme.needs_renewal": MagicMock(return_value=True),
                "acme.info": MagicMock(
                    side_effect=[{"name": "old cert"}, {"name": "new cert"}]
                ),
                "acme.cert": MagicMock(
                    return_value={"result": True, "comment": "Mockery"}
                ),
            },
        ):
            self.assertEqual(
                acme.cert("test"),
                {
                    "name": "test",
                    "result": True,
                    "comment": ["Mockery"],
                    "changes": {
                        "old": {"name": "old cert"},
                        "new": {"name": "new cert"},
                    },
                },
            )
